/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE EXPRESSION PARSER
 *
 *	Parses an expression, and builds an extended parse tree.
 *
 *	Public functions:
 *		MKTABLE()	Parse/create a TABLE
 *		MKSTRING()	Parse/create a STRING
 *		EXPR()		Parse an expression and create
 *				an extended parse tree
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "scan.h"
#include "token.h"
#include "glue.h"

#include "proto.h"


struct node	*newnode();
struct node	*expr();
struct syment	*findsym();
struct node	*link();
struct node	*cut();


/*
 *	MKTABLE()
 *	Purpose:	Parse a TABLE, starting at first element, and
 *			create it.
 *	Receives:	-
 *	Returns:	The address of the first table element.
 *
 *	WARNING:	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 *			Nesting tables by more than two levels
 *			may break the parser (stack overflow).
 *	WARNING:	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 */
struct node *mktable(ap)
int	*ap;
{
	static int	tbl[TBLLEN], rel[TBLLEN];
	struct syment	*sym;
	char	sname[NAMELEN+1];
	int	i, j;
	struct node	*root, *t = NULL;

	for (i=0;;) {
		if (i >= TBLLEN) error(ETBLL, FATAL);
		if (Token == ADDROF) {
			next();
			strncpy(sname, Strval, NAMELEN);
			sname[NAMELEN] = 0;
			symname();
			if ((sym = findsym(sname, 0)) == NULL)
				if ((sym = findsym(sname, 1)) == NULL)
					sym = undeclared(sname);
			sym->flags |= REFERD;
			if (sym->flags & RVALUE) error(ELVAL, ERROR);
			if (sym->flags & EXTERN) error(EXADDR, ERROR);
			if (!(sym->flags & STATIC)) error(EAADDR, ERROR);
			tbl[i] = sym->addr;
			rel[i++] = DATA;
		}
		else if (Token == STRING) {
			tbl[i] = Dtop;
			t = link(t, mkstring());
			rel[i++] = DATA;
		}
		else if (Token == LBRACK) {
			next();
			t = link(t, mktable(&tbl[i]));
			rel[i++] = DATA;
		}
		else {
			tbl[i] = cexpr();
			rel[i++] = NONE;
		}
		if (Token != COMMA) break;
		next();
	}
	match(RBRACK, "]", ERROR);

	root = newnode(G_INI);
	root->v.val = i;
	root = prepend(root, G_NUM);
	*ap = root->v.val = Dtop;
	for (j=0; j<i; j++) {
		if (rel[j] != NONE)
			root = prepend(root, (rel[j] == CODE)? G_RLA: G_REL);
		root = prepend(root, G_NUM);
		root->v.val = tbl[j];
	}
	alloc(i, 1);

	if (t) {
		t = cut(t);
		root = link(t, root);
	}
	return(root);
}


/*
 *	MKSTRING()
 *	Purpose:	Parse and create a STRING literal.
 *	Receives:	-
 *	Returns:	The address of the first character in the string.
 */
struct node *mkstring() {
	int	i, k;
	struct node	*root;

	k = strlen(Strval);
	root = newnode(G_INI);
	root->v.val = k+1;
	root = prepend(root, G_NUM);
	root->v.val = Dtop;
	for (i=0; i<k; i++) {
		root = prepend(root, G_NUM);
		root->v.val = Strval[i] & 0xff;
	}
	root = prepend(root, G_NUM);
	root->v.val = 0;
	alloc(k+1, 1);
	next();
	return(root);
}

/*
 *	ORDER(EXP, INI)
 *	Purpose:	Extract INI nodes from an expression
 *	Receives:	EXP	- Pointer to the expression tree
 *				to scan
 *			INI	- Pointer to the root node of the
 *				tree where INI nodes will be
 *				prepended
 *	Returns:	A parse for the function call.
 */
void order(exp, inip)
struct node	*exp, **inip;
{
	if (!exp) return;
	if (exp->token == G_LA && exp->left) {
		*inip = link(*inip, exp->left);
		exp->left = NULL;
		return;
	}
	order(exp->left, inip);
	order(exp->right, inip);
}


/*
 *	FNCALL(NODE)
 *	Purpose:	Parse a function call
 *	Receives:	NODE	- A pointer to the subtree that computes
 *				the function address.
 *	Returns:	A parse for the function call.
 */
struct node *fncall(node)
struct node	*node;
{
	struct node	*root, *t, *ini = NULL;
	int	narg = 0;

	next();
	root = t = NULL;
	if (Token != RPAREN) for (;;) {
		narg++;
		if (t) {
			t->left = newnode(G_PSH);
			t = t->left;
		}
		else {
			root = t = newnode(G_PSH);
		}
		t->right = expr(0);
		order(t->right, &ini);
		if (Token != COMMA) break;
		next();
	}
	if (ini) ini = cut(ini);
	rparen();
	fetch(&node);
	root = link(ini, root);
	root = link(root, node);
	root->token = G_VCA;
	t = newnode(G_RMA);
	t->v.val = narg;
	t->left = root;
	return(t);
}


/*
 *	PRIMARY()
 *	Purpose:	Parse a primary factor.
 *			PRIMARY() handles numeric literals, SYMBOLS,
 *			structures, STRINGS, TABLES, and grouping.
 *	Receives:	-
 *	Returns:	A parse tree
 */
struct node *primary() {
	struct node	*node;
	struct syment	*sym;
	int	addr;

	switch (Token) {
		case NUMBER:
			node = newnode(G_LN);
			node->v.val = Value;
			node->flags = CONST;
			next();
			break;

		case SYMBOL:
			node = newnode(G_SYM);
			if ((sym = findsym(Strval, 0)) != NULL) {
				node->flags = 0;
			}
			else if ((sym = findsym(Strval, 1)) != NULL) {
				node->flags = GLOBAL;
			}
			else {
				sym = undeclared(Strval);
			}
			sym->flags |= REFERD;
			if (sym->flags & RVALUE) {
				node->token = G_LN;
				node->v.val = (sym->flags & STRUCT)?
				stcsize(sym): sym->addr;
				node->flags |= CONST;
			}
			else {
				node->v.sym = sym;
				node->flags |= DIRECT;
			}
			next();
			break;

		case STRING:
			node = newnode(G_LA);
			node->v.val = Dtop;
			node->left = mkstring();
			node->flags = CONST;
			break;

		case LPAREN:
			next();
			node = expr(1);
			node->flags = CONST;
			rparen();
			break;

		case LBRACK:
			next();
			node = newnode(G_LA);
			node->left = mktable(&addr);
			node->v.val = addr;
			node->flags = CONST;
			break;

		default:
			expect("factor", SEVERE);
	}

	return(node);
}


/*
 *	STRUCT_XS()
 *	Purpose:	Parse a binary indirection.
 *	Receives:	-
 *	Returns:	A parse tree.
 */
struct node *struct_xs() {
	struct node	*root, *t;

	root = primary();
	if (Token == LPAREN) root = fncall(root);
	while (Token == LV) {
		next();
		fetch(&root);
		t = newnode(G_ADD);
		t->left = root;
		root = t;
		root->right = primary();
		fetch(&root->right);
		t = newnode(G_LV);
		t->left = root;
		root = t;
		root->flags = 0;
		if (Token == LPAREN) root = fncall(root);
	}
	return(root);
}


/*
 *	FACTOR()
 *	Purpose:	Parse a factor. This includes the handling of
 *			the unary operators . - + ! ~ @.
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *factor() {
	struct node	*root;

	switch (Token) {
		case LV:
			root = newnode(G_LV);
			next();
			root->left = factor();
			fetch(&root->left);
			root->flags = 0;
			break;

		case SUB:
		case NOT:
		case LNOT:
			root = newnode((Token == SUB)? G_NEG:
			(Token == NOT)? G_COM: G_NOT);
			next();
			root->left = factor();
			fetch(&root->left);
			root->flags = CONST;
			break;

		case ADDROF:
			next();
			root = primary();
			if (!(root->flags & DIRECT)) {
				error(ELVAL, ERROR);
				break;
			}
			if (root->v.sym->flags & EXTERN) {
				error(EXADDR, ERROR);
				break;
			}
			if (root->v.sym->flags & STATIC) {
				root->token = G_LA;
				root->v.val = root->v.sym->addr;
			}
			else {
				struct node	*t;

				root->token = G_LN;
				root->v.val = root->v.sym->addr;
				t = newnode(G_ADJ);
				t->left = root;
				root = t;
			}
			break;

		case ADD:
			next(); root = factor(); break;

		default:
			root = struct_xs();
	}

	return(root);
}


/*
 *	TERM()
 *	Purpose:	Parse a term (operators: * / %).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *term() {
	struct node	*root, *t;

	root = factor();
	for (;;) {
		if (Token == MUL || Token == DIV || Token == MOD) {
			t = newnode((Token == MUL)? G_MUL:
			(Token == DIV)? G_DIV: G_MOD);
			next();
			t->left = root;
			t->right = factor();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	SUM()
 *	Purpose:	Parse a sum (operators: + -).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *sum() {
	struct node	*root, *t;

	root = term();
	for (;;) {
		if (Token == ADD || Token == SUB) {
			t = newnode((Token == ADD)? G_ADD: G_SUB);
			next();
			t->left = root;
			t->right = term();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	SH_ROT()
 *	Purpose:	Parse a chain of shift/rotate operations
 *			(operators: << >> {{ }}).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *sh_rot() {
	struct node	*root, *t;
	int	op;

	root = sum();
	for (;;) {
		if (Token == SHL || Token == SHR || Token == ROL ||
		Token == ROR) {
			switch(Token) {
				case SHL:	op = G_SHL; break;
				case SHR:	op = G_SHR; break;
				case ROL:	op = G_ROL; break;
				case ROR:	op = G_ROR; break;
			}
			t = newnode(op);
			next();
			t->left = root;
			t->right = sum();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	COMP()
 *	Purpose:	Parse a chain of comparisons (operators: < > <= >=).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *comp() {
	struct node	*root, *t;
	int	op, u;

	root = sh_rot();
	for (;;) {
		if (Token == KUNSIGN) {
			next(); u = 1;
		}
		else
			u = 0;

		if (Token == GRTR || Token == LESS || Token == GTEQ ||
		Token == LSEQ) {
			switch(Token) {
				case GRTR:	op = u? G_UGT: G_GT; break;
				case LESS:	op = u? G_ULT: G_LT; break;
				case GTEQ:	op = u? G_UGE: G_GE; break;
				case LSEQ:	op = u? G_ULE: G_LE; break;
			}
			t = newnode(op);
			next();
			t->left = root;
			t->right = sh_rot();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	EQUATION()
 *	Purpose:	Parse a chain of comparisons (operators: == !=).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *equation() {
	struct node	*root, *t;

	root = comp();
	for (;;) {
		if (Token == EQU || Token == NEQ) {
			t = newnode((Token == EQU)? G_EQ: G_NE);
			next();
			t->left = root;
			t->right = comp();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	BINAND()
 *	Purpose:	Parse a chain of binary AND operations (&).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *binand() {
	struct node	*root, *t;

	root = equation();
	for (;;) {
		if (Token == AND) {
			t = newnode(G_AND);
			next();
			t->left = root;
			t->right = equation();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	BINXOR()
 *	Purpose:	Parse a chain of binary XOR operations (^).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *binxor() {
	struct node	*root, *t;

	root = binand();
	for (;;) {
		if (Token == XOR) {
			t = newnode(G_XOR);
			next();
			t->left = root;
			t->right = binand();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	BINOR()
 *	Purpose:	Parse a chain of binary OR operations (|).
 *	Receives:	-
 *	Returns:	a parse tree
 */
struct node *binor() {
	struct node	*root, *t;

	root = binxor();
	for (;;) {
		if (Token == OR) {
			t = newnode(G_OR);
			next();
			t->left = root;
			t->right = binxor();
			root = t;
			fetch(&root->left);
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	LOGAND()
 *	Purpose:	Parse a chain of logical AND operations (&&).
 *			Create code for short-circuit evaluation.
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *logand() {
	struct node	*root, *t;
	int	dest = -1;

	root = binor();
	for (;;) {
		if (Token == LAND) {
			fetch(&root);
			next();
			if (dest == -1) {
				t = newnode(G_LBL);
				dest = t->v.val = newlabel();
				t->right = root;
				root = t;
				root->flags = CONST;
			}
			root->token = G_JFA;
			t = newnode(G_LBL);
			t->v.val = dest;
			t->left = root;
			root = t;
			root->right = binor();
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	LOGOR()
 *	Purpose:	Parse a chain of logical OR operations (||).
 *			Create code for short-circuit evaluation.
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *logor() {
	struct node	*root, *t;
	int	dest = -1;

	root = logand();
	for (;;) {
		if (Token == LOR) {
			fetch(&root);
			next();
			if (dest == -1) {
				t = newnode(G_LBL);
				dest = t->v.val = newlabel();
				t->right = root;
				root = t;
				root->flags = CONST;
			}
			root->token = G_JTR;
			t = newnode(G_LBL);
			t->v.val = dest;
			t->left = root;
			root = t;
			root->right = logand();
			fetch(&root->right);
			root->flags = CONST;
		}
		else
			return(root);
	}
}


/*
 *	COND()
 *	Purpose:	Parse a conditional expression
 *			( ternary operator -> , ).
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *cond() {
	struct node	*root, *t;
	int	false, end;

	root = logor();
	if (Token == ARROW) {
		fetch(&root);
		next();

		t = newnode(G_JFA);
		false = t->v.val = newlabel();
		t->left = root;
		root = t;

		t = newnode(G_JMP);
		end = t->v.val = newlabel();
		t->left = root;
		root = t;

		root->right = cond();
		fetch(&root->right);
		match(COMMA, "','", ERROR);

		t = newnode(GLUE);
		t->left = root;
		root = t;

		root->right = newnode(G_LBL);
		t = root->right;

		t->v.val = end;
		t->left = newnode(G_LBL);
		t->left->v.val = false;
		t->right = cond();
		fetch(&t->right);

		root->flags = CONST;
	}

	return(root);
}


/*	BCODE instructions for use with combined assignment operators */
static int ops[] = {
	G_ROR, G_ROL, G_SHR, G_SHL, G_XOR, G_OR, G_AND,
	G_MOD, G_DIV, G_MUL, G_SUB, G_ADD
};

/*
 *	ASSIGNMENT()
 *	Purpose:	Parse an assignment
 *			(operators: = += -= *= /= %= &= |= ^= <<= >>= {{= }}=).
 *	Receives:	-
 *	Returns:	an extended parse tree.
 */
struct node *assignment() {
	struct node	*root, *t;
	int	op = 0;

	root = cond();
	switch (Token) {
		case MOV:
			if (root->flags & CONST) error(ELVAL, ERROR);

			next();

			if (root->flags & DIRECT) {
				if (root->v.sym->flags & EXTERN) {
					root->token = G_SX;
					root->v.val = findsx(root->v.sym->name);
				}
				else {
					root->token = (root->flags & GLOBAL ||
					root->v.sym->flags & STATIC)?
					G_SG: G_SL;
					root->v.val = root->v.sym->addr;
				}
				root->left = assignment();
				fetch(&root->left);
			}
			else {
				if (root->token == G_LV) {
					t = root;
					root = root->left;
					freenode(t);
				}
				t = newnode(G_SV);
				t->left = root;
				root = t;
				root->right = assignment();
				fetch(&root->right);
			}

			break;

		case AADD:	op++;
		case ASUB:	op++;
		case AMUL:	op++;
		case ADIV:	op++;
		case AMOD:	op++;
		case AAND:	op++;
		case AOR:	op++;
		case AXOR:	op++;
		case ASHL:	op++;
		case ASHR:	op++;
		case AROL:	op++;
		case AROR:
			if (root->flags & CONST) error(ELVAL, ERROR);

			next();

			if (root->flags & DIRECT) {
				struct syment	*lv;
				int	globl, sx;

				lv = root->v.sym;
				globl = (root->flags&GLOBAL) != 0;

				if (lv->flags & EXTERN) {
					root->token = G_LX;
					sx = root->v.val = findsx(lv->name);
				}
				else {
					root->token = (globl ||
					lv->flags & STATIC)?  G_LG: G_LL;
					root->v.val = root->v.sym->addr;
				}
				t = newnode(ops[op]);
				t->left = root;
				t->right = assignment();
				fetch(&t->right);
				root = t;

				if (lv->flags & EXTERN) {
					t = newnode(G_SX);
					t->v.val = sx;
				}
				else {
					t = newnode((globl ||
					lv->flags & STATIC)? G_SG: G_SL);
					t->v.val = lv->addr;
				}
				t->left = root;
				root = t;
			}
			else {
				if (root->token == G_LV) {
					t = root;
					root = root->left;
					freenode(t);
				}
				t = newnode(G_SAV);
				t->right = root;
				root = t;
				t = newnode(G_LV);
				t->left = root;
				root = t;

				t = newnode(ops[op]);
				t->left = root;
				root = t;

				root->right = assignment();
				fetch(&root->right);
				t = newnode(G_SV);
				t->left = root;
				root = t;
			}

			break;
	}

	return(root);
}


/*
 *	EXPR(COMMA)
 *	Purpose:	Parse an expression, and create an extended parse
 *			tree.
 *	Receives:	COMMA	- Comma-operator allowed.
 *	Returns:	An extended parse tree.
 */
struct node *expr(comma) {
	struct node	*root, *t;
	int	first = 1;

	root = assignment();
	fetch(&root);

	if (comma) for (;;) {
		if (Token != COMMA) break;
		next();
		if (first) {
			t = newnode(GLUE);
			t->left = root;
			root = t;
		}
		t = assignment();
		fetch(&t);
		root->token = G_DRO;
		root = link(root, t);
	}

	return(root);
}
