/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE OPTIMIZER
 *
 *	The optimizer performs some transformations on the parse tree
 *	to make the resulting code more compact, and faster.
 *
 *	Public functions:
 *		OPTIMIZE()	Optimize the parse tree
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "glue.h"

#include "proto.h"


/*
 *	HEAD(LI)
 *	Purpose:	Find the head of a label node. Skip labels,
 *			GLUE nodes and empty UNV-instructions between
//			the label and the head.
 *	Receives:	LI	- the pool index of the label
 *	Returns:	The pool index of the head. When there is no
 *			parent node, the head is LI.
 */
head(li) {
	struct node	*a = &Nodes[li];
	int	i;

	for (i=0; i<Nsize; i++) {
		if (Nodes[i].left == a && Nodes[i].right == NULL ||
		Nodes[i].right == a) {
			if (Nodes[i].token == GLUE || Nodes[i].token == G_LBL
			|| Nodes[i].token == G_UNV && !Nodes[i].v.val)
				return(head(i));
			else
				return(i);
		}
	}
	return(li);
}


/*
 *	FOLLOW(SI)
 *	Purpose:	Follows subsequent JMP's thru the node pool.
 *			Used to find the destination of a jump chain.
 *	Receives:	SI	- The pool index of the origin (source)
 *	Returns:	The pool index of the final destination
 */
follow(si) {
	int	i, l, ni;

	for (;;) {
 		l = Nodes[si].v.val;
		for (i=0; i<Nsize; i++) {
			if (Nodes[i].token == G_LBL && Nodes[i].v.val == l) {
				ni = head(i);
				break;
			}
		}
		if (Nodes[ni].token != G_JMP) return(si);
		si = ni;
	}
}


/*
 *	OPT_JUMPS()
 *	Purpose:	Elimininates jump to jumps. Directly modifies the
 *			extended parse tree.
 *	Receives:	-
 *	Returns:	-
 */
void opt_jumps() {
	struct node	*f, *tn;
	int	i, j, t, dl;

	for (i=0; i<Nsize; i++) {
		t = Nodes[i].token;
		if (t == G_JMP || t == G_JTR || t == G_JFA) {
			Nodes[i].v.val = Nodes[follow(i)].v.val;

			dl = Nodes[i].v.val;
			for (j=0; j<Nsize; j++) {
				if (Nodes[j].token == G_LBL &&
				Nodes[j].v.val == dl) {
					tn = &Nodes[j];
					for (;;) {
						f = tn->right;
						if (!f) f = tn->left;
						if ((tn = f) == NULL) break;
						if (f->token == G_JMP &&
						f->v.val == dl) {
							f->token = GLUE; break;
						}
						else if (f->token != GLUE) {
							break;
						}
					}
				}
			}
		}
	}
}


/*
 *	INSWI(LID, DIST)
 *	Purpose:	Checks if a label is referenced in a SWITCH.
 *	Receives:	LID	- Label to check
 *			DIST	- Distance between SWI and end of switch.
 *	Returns:	== 0 -> label not referenced
 *			!= 0 -> label referenced
 */
inswi(lid, dist) {
	struct node	*np;
	int	i, k, found = 0;

	dist = dist*2 + 3;
	for (i=0; i<Nsize; i++) {
		if (Nodes[i].token == G_LBL) {
			np = Nodes[i].right;
			for (k=1; k<dist; k++) {
				if (!np || np->token != G_NUM) break;
				if ((k&1) && np->v.val == lid) found = 1;
				np = np->right;
			}
			if (np->token == G_SWI) {
				return(found);
			}
		}
	}
	return(0);
}


/*
 *	OPT_LABELS()
 *	Purpose:	Removes unused labels. Directly modifies the
 *			extended parse tree.
 *	Receives:	-
 *	Returns:	-
 */
void opt_labels() {
	int	i, j, t;

	for (i=0; i<Nsize; i++) {
		if (Nodes[i].token == G_LBL) {
			for (j=0; j<Nsize; j++) {
				t = Nodes[j].token;

				if ((t == G_JMP || t == G_JTR || t == G_JFA) &&
				Nodes[j].v.val==Nodes[i].v.val)
					break;

				if (t == G_NUM && Nodes[j].left &&
				Nodes[j].left->token == G_RLA)
					break;

				if (t == G_SWI &&
				inswi(Nodes[i].v.val, Nodes[j].v.val))
					break;
			}

			if (j>=Nsize) {
				for (j=0; j<Forwp; j++) {
					if (Forwtab[j] == Nodes[i].v.val)
						break;
				}
				if (j >= Forwp) Nodes[i].token = GLUE;
			}
		}
	}
}


/*
 *	OPT_BRANCHES()
 *	Purpose:	Eliminates conditional branches with constant
 *			conditions. Directly modifies the extended
 *			parse tree.
 *	Receives:	-
 *	Returns:	-
 */
void opt_branches() {
	struct node	**head;
	int	i, t;

	for (i=0; i<Nsize; i++) {
		t = Nodes[i].token;
		if (t == G_JTR || t == G_JFA) {
			if (Nodes[i].right)
				head = &Nodes[i].right;
			else
				head = &Nodes[i].left;
			if ((*head)->token == G_LN) {
				if ((*head)->v.val) {
					Nodes[i].token = (t==G_JTR)? G_JMP:
					GLUE;
				}
				else {
					Nodes[i].token = (t==G_JFA)? G_JMP:
					GLUE;
				}
				*head = NULL;
			}
		}
	}
}


/*
 *	RC_OPT(ROOT)
 *	Purpose:	Removes superflous instructions from right sides
 *			of sub-expressions (like +0, *1, &0).
 *	Receives:	ROOT	- A pointer to the parse tree.
 *	Returns:	A pointer to the optimized parse treeA.
 */
struct node *rc_opt(root)
struct node	*root;
{
	struct node	*rc = root->right;
	uint	log, v;

	if (rc->v.val == 0) {
		switch (root->token) {
			case G_ADD: case G_SUB:
			case G_SHL: case G_SHR:
			case G_ROL: case G_ROR:
			case G_XOR: case G_OR:
				root->token = GLUE;
				root->right = NULL;
				break;

			case G_MUL: case G_AND:
				root->token = G_LN;
				root->v.val = 0;
				root->left = root->right = NULL;
				break;

			case G_DIV:
				/* divide by zero */
				break;
		}
		return(root);
	}

	if (rc->v.val == 1) {
		switch (root->token) {
			case G_MUL: case G_DIV:
				root->token = GLUE;
				root->right = NULL;
				break;

			case G_MOD:
				root->token = G_LN;
				root->v.val = 0;
				root->left = root->right = NULL;
				break;
		}
		return(root);
	}

	if (rc->v.val == -1) {
		switch (root->token) {
			case G_AND:
				root->token = GLUE;
				root->right = NULL;
				break;

			case G_XOR:
				root->token = G_COM;
				root->right = NULL;
				break;

			case G_OR:
				root->token = G_LN;
				root->v.val = -1;
				root->left = root->right = NULL;
				break;
		}
		return(root);
	}

	v = 2; log = 1;
	while (log<_BITS_) {
		if (rc->v.val == v) break;
		v *= 2;
		log++;
	}
	if (rc->v.val == v) {
		switch (root->token) {
			case G_MUL:
				root->token = G_SHL;
				rc->v.val = log;
				break;

			case G_DIV:
				root->token = G_SHR;
				rc->v.val = log;
				break;
		}
	}

	return(root);
}


/*
 *	LC_OPT(ROOT)
 *	Purpose:	Removes superflous instructions from left sides
 *			of sub-expressions (like 0+, 1*, 0&).
 *	Receives:	ROOT	- A pointer to the parse tree.
 *	Returns:	A pointer to the optimized parse treeA.
 */
struct node *lc_opt(root)
struct node	*root;
{
	struct node	*lc = root->left;
	uint	log, v;

	if (lc->v.val == 0) {
		switch (root->token) {
			case G_MUL: case G_AND:
			case G_SHL: case G_SHR:
			case G_ROL: case G_ROR:
				root->token = G_LN;
				root->v.val = 0;
				root->left = root->right = NULL;
				break;

			case G_SUB:
				root->token = G_NEG;
				root->left = NULL;
				break;

			case G_ADD: case G_XOR: case G_OR:
				root->token = GLUE;
				root->left = NULL;
				break;
		}
		return(root);
	}

	if (lc->v.val == 1) {
		switch (root->token) {
			case G_MUL:
				root->token = GLUE;
				root->left = NULL;
				break;
		}
		return(root);
	}

	if (lc->v.val == -1) {
		switch (root->token) {
			case G_AND:
				root->token = GLUE;
				root->left = NULL;
				break;

			case G_XOR:
				root->token = G_COM;
				root->left = NULL;
				break;

			case G_OR:
				root->token = G_LN;
				root->v.val = -1;
				root->left = root->right = NULL;
				break;
		}
		return(root);
	}

	v = 2; log = 1;
	while (log<_BITS_) {
		if (lc->v.val == v) break;
		v *= 2;
		log++;
	}
	if (lc->v.val == v) {
		switch (root->token) {
			case G_MUL:
				root->token = G_SHL;
				lc->v.val = log;
				break;
		}
	}

	return(root);
}


/*
 *	OPT_OPER(ROOT)
 *	Purpose:	Remove superflous operations from expressions.
 *	Receives:	ROOT	- A pointer to the extended parse tree.
 *	Returns:	A pointer to the optimized parse tree.
 */
struct node *opt_oper(root)
struct node	*root;
{
	struct node	*lc, *rc;

	if ((lc = root->left) == NULL) return(root);
	if ((rc = root->right) == NULL) return(root);

	if (lc->token == G_LN) root = lc_opt(root);
	if (rc->token == G_LN) root = rc_opt(root);

	return(root);
}


/*
 *	OPT_CEXPR(ROOT)
 *	Purpose:	Evaluate constant sub-expressions, and remove
 *			superflous operations.
 *	Receives:	ROOT	- A pointer to the extended parse tree.
 *	Returns:	A pointer to the optimized tree.
 */
struct node *opt_cexpr(root)
struct node	*root;
{
	struct node	*lc, *rc;
	int	v, lv, rv;
	uint	ulv, urv;

	if (!root) return(root);
	if (root->token == G_NUM) return(root);	/* Dont follow N-lists! */

	root->left = opt_cexpr(root->left);
	root->right = opt_cexpr(root->right);

	lc = root->left;
	if ((rc = root->right) != NULL) {
		if (lc->token != G_LN || rc->token != G_LN)
			return(opt_oper(root));
	}
	else {
		if (lc->token != G_LN) return(opt_oper(root));
	}

	ulv = lv = lc->v.val;
	urv = rv = rc->v.val;
	switch(root->token) {
		case G_ADD:	v = lv + rv; break;
		case G_SUB:	v = lv - rv; break;
		case G_MUL:	v = ulv * urv; break;
		case G_DIV:	v = ulv / urv; break;
		case G_MOD:	v = ulv % urv; break;
		case G_NEG:	v = -lv; break;

		case G_AND:	v = lv & rv; break;
		case G_OR:	v = lv | rv; break;
		case G_XOR:	v = lv ^ rv; break;
		case G_SHL:	v = ulv << urv; break;
		case G_SHR:	v = ulv >> urv; break;
		case G_ROL:	v = rol(lv, rv); break;
		case G_ROR:	v = ror(lv, rv); break;
		case G_COM:	v = ~lv; break;

		case G_EQ:	v = (lv == rv)? -1: 0; break;
		case G_NE:	v = (lv != rv)? -1: 0; break;
		case G_LT:	v = (lv < rv)? -1: 0; break;
		case G_GT:	v = (lv > rv)? -1: 0; break;
		case G_LE:	v = (lv <= rv)? -1: 0; break;
		case G_GE:	v = (lv >= rv)? -1: 0; break;
		case G_ULT:	v = (ulv < urv)? -1: 0; break;
		case G_UGT:	v = (ulv > urv)? -1: 0; break;
		case G_ULE:	v = (ulv <= urv)? -1: 0; break;
		case G_UGE:	v = (ulv >= urv)? -1: 0; break;
		case G_NOT:	v = lv? 0: -1; break;

		default:	return(root);
	}

	root->token = G_LN;
	root->v.val = v;
	root->right = root->left = NULL;

	return(root);
}


/*
 *	OPTIMIZE(ROOT)
 *	Purpose:	Optimizes the generated parse tree, using
 *			the following steps:
 *				1) Evaluation of constant expressions
 *				2) Elimination of conditional branches
 *				   with constant conditions
 *				3) Elimination of jumps to jumps.
 *	Receives:	ROOT	- A pointer to the extended parse tree.
 *	Returns:	A pointer to the optimized tree.
 */
struct node *optimize(root)
struct node	*root;
{
	root = opt_cexpr(root);
	opt_branches();
	opt_jumps();
/*	opt_labels();	*/
	return(root);
}
