/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE SCANNER
 *
 *	The scanner reads the source program, and creates a token stream
 *	for the parser.
 *
 *	Public functions:
 *		NEXT()		Read next token
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include "token.h"
#include "scan.h"
#include "defs.h"
#include "data.h"
#include "error.h"

#include "proto.h"


static int	Lastrd;				/* Result of last read */


#define isoct(c)	('0' <= (c) && (c) <= '7')


/*
 *	SC_INIT()
 *	Purpose:	Initialize the scanner.
 *	Receives:	FNAME	- The name of the input file
 *	Returns:	-
 */
void sc_init(fname)
char	*fname;
{
	Buffer[0] = Bp = Btop = 0;
	Lastrd = 0;
	if (fname == NULL) {
		Infile = fileno(stdin);
		strcpy(Inname, "(SYSIN)");
	}
	else {
		if ((Infile = open(fname, O_RDONLY)) < 0)
			error(EINPUT, FATAL);
	}
	setmode(Infile, O_BINARY);
}


/*
 *	READB()
 *	Purpose:	Read a block from the source file.
 *	Receives:	-
 *	Returns:	0  - everything fine
 *			-1 - EOF reached
 */
readb() {
	if (Lastrd < 0) {
		return((Bp >= Btop)? -1: 0);
	}
	if (Bp != Btop && Btop > 0)
		memcpy(Buffer, &Buffer[Bp], Btop-Bp);
	Bp = Btop-Bp;
	if ((Btop = read(Infile, &Buffer[Bp], BUFLEN)) < 0)
		error(EIO, FATAL);
	if (Btop < BUFLEN) Lastrd = -1;
	Btop += Bp;
	Bp = 0;
	Buffer[Btop] = 0;
	return((!Btop)? -1: 0);
}


/*
 *	COMMENT()
 *	Purpose:	Skip a comment.
 *	Receives:	-
 *	Returns:	0  - end of comment found (OK)
 *			-1 - EOF found in comment
 */
comment() {
	rch();
	rch();
	while (ch() != '\n') {
		if (!ch()) {
			if (Bp >= Btop-TOKLEN) {
				if (readb() < 0) return(-1);
			}
		}
		rch();
	}
	return(0);
}


/*
 *	SKIP()
 *	Purpose:	Skip white space and comments.
 *			Read next block, if necessary.
 *	Receives:	-
 *	Returns:	0  - everything fine
 *			-1 - EOF found
 */
skip() {
	for (;;) {
		if (Bp >= Btop-TOKLEN) {
			if (readb() < 0) return(-1);
		}
		if (ch() == '/') {
			if (nch() == '/') {
				if (comment() < 0) return(-1);
			}
		}
		if (!iswhite(ch())) break;
		if (ch() == '\n') Lno++;
		rch();
	}
	return(0);
}


/*
 *	RCHE()
 *	Purpose:	Read a (possibly escaped) character.
 *			Convert escape sequences to ASCII values.
 *	Receives:	-
 *	Returns:	The ASCII value of the character.
 */
rche() {
	int	v, i;

	if (ch() != '\\') return(rch());

	rch();
	switch(ch()) {
		case 'A': case 'a':	rch(); return('\a');
		case 'B': case 'b':	rch(); return('\b');
		case 'E': case 'e':	rch(); return('\033');
		case 'F': case 'f':	rch(); return('\f');
		case 'N': case 'n':	rch(); return('\n');
		case 'R': case 'r':	rch(); return('\r');
		case 'S': case 's':	rch(); return(' ');
		case 'T': case 't':	rch(); return('\t');
		case 'V': case 'v':	rch(); return('\v');
		default:	if (isoct(ch())) {
					for (v=i=0; i<3 && isoct(ch()); i++)
						v = v*8 + (rch()-'0');
					return(v);
				}
				return(rch());
	}
}


/*
 *	KEYWORD(S)
 *	Purpose:	Check whether a string is a keyword.
 *	Receives:	S	- the string
 *	Returns:	The token associated with the keyword,
 *			or a special failure value, if the string
 *			does not contain a keyword.
 */
keyword(s)
char	*s;
{
	char	*p;

	for (p=s; *p; p++)
		if (islower(*p))
			*p = toupper(*p);

	switch(*s) {
		case 'B':
			if (!strcmp(s, "BREAK")) return(KBREAK);
			break;

		case 'C':
			if (!strcmp(s, "CASE")) return(KCASE);
			break;

		case 'D':
			if (!strcmp(s, "DEFAULT")) return(KDFLT);
			else if (!strcmp(s, "DO")) return(KDO);
			break;

		case 'E':
			if (!strcmp(s, "ELSE")) return(KELSE);
			else if (!strcmp(s, "EXTERN")) return(KEXTERN);
			break;

		case 'F':
			if (!strcmp(s, "FOR")) return(KFOR);
			if (!strcmp(s, "FORWARD")) return(KFORWD);
			break;

		case 'H':
			if (!strcmp(s, "HALT")) return(KHALT);
			break;

		case 'I':
			if (!strcmp(s, "IF")) return(KIF);
			break;

		case 'L':
			if (!strcmp(s, "LOOP")) return(KLOOP);
			break;

		case 'O':
			if (!strcmp(s, "OPTIONS")) return(KOPTS);
			break;

		case 'P':
			if (!strcmp(s, "PUBLIC")) return(KPUBLIC);
			break;

		case 'R':
			if (!strcmp(s, "RETURN")) return(KRETURN);
			break;

		case 'S':
			if (!strcmp(s, "STATIC")) return(KSTATIC);
			else if (!strcmp(s, "STRUCTURE")) return(KSTRUCT);
			else if (!strcmp(s, "SWITCH")) return(KSWITCH);
			break;

		case 'U':
			if (!strcmp(s, "U")) return(KUNSIGN);
			else if (!strcmp(s, "UFOR")) return(KUFOR);
			else if (!strcmp(s, "UNSIGNED")) return(KUNSIGN);
			else if (!strcmp(s, "UNTIL")) return(KUNTIL);
			break;

		case 'V':
			if (!strcmp(s, "VAR")) return(KVAR);
			else if (!strcmp(s, "VALUE")) return(KVALUE);
			break;

		case 'W':
			if (!strcmp(s, "WHILE")) return(KWHILE);
			break;
	}

	return(FUCKUP);
}


/*
 *	SCANVAL()
 *	Purpose:	Scan a numeric literal, auto-detect radix.
 *			No overflow checking is done !!!
 *	Receives:	-
 *	Returns:	The value of the literal.
 */
scanval() {
	int	radix, c, v, i=0;
	char	*digits, p;
	int	d, o, x;

	digits = "0123456789ABCDEF";

	if (ch() == '0') {
		rch();
		if (ch() == 'x' || ch() == 'X') {
			rch(); radix = 16;
		}
		else if (ch() == 'b' || ch() == 'B') {
			rch(); radix = 2;
		}
		else {
			radix = 8;
		}
	}
	else
		radix = 10;

	p = digits[radix];
	digits[radix] = 0;

	x = 1; v = 0;
	while (strchr(digits, c = toupper(ch()))) {
		if (++i == TOKLEN) error(ETOKL, FATAL);
		o = v;
		d = c - (isdigit(c)? '0': ('A'-10));
		v = v*radix + d;
		if ((v-d)/radix != o && x) {
			error(EOVFL, WARNING); x = 0;
		}
		rch();
	}
	back();

	digits[radix] = p;

	return(v);
}


/*
 *	SCAN()
 *	Purpose:	Extract a token from the source file.
 *	Receives:	-
 *	Returns:	The extracted token.
 */
scan() {
	Strval[0] = Value = 0;

	if (Btop < 0) return(ENDFILE);
	if (ch()) rch();
	if (skip() < 0) {
		Btop = -1;
		return(ENDFILE);
	}

	if (ch() == '.') return(LV);
	else if (ch() == '@') return(ADDROF);
	else if (ch() == ';') return(SEMI);
	else if (ch() == ',') return(COMMA);
	else if (ch() == '~') return(NOT);
	else if (ch() == '(') return(LPAREN);
	else if (ch() == ')') return(RPAREN);
	else if (ch() == '[') return(LBRACK);
	else if (ch() == ']') return(RBRACK);
	else if (ch() == ':') return(COLON);
	else if (ch() == '=') {
		if (nch() == '=') {
			rch(); return(EQU);
		}
		return(MOV);
	}
	else if (ch() == '!') {
		if (nch() == '=') {
			rch(); return(NEQ);
		}
		return(LNOT);
	}
	else if (ch() == '+') {
		if (nch() == '=') {
			rch(); return(AADD);
		}
		return(ADD);
	}
	else if (ch() == '-') {
		if (nch() == '=') {
			rch(); return(ASUB);
		}
		else if (nch() == '>') {
			rch(); return(ARROW);
		}
		return(SUB);
	}
	else if (ch() == '*') {
		if (nch() == '=') {
			rch(); return(AMUL);
		}
		return(MUL);
	}
	else if (ch() == '/') {
		if (nch() == '=') {
			rch(); return(ADIV);
		}
		return(DIV);
	}
	else if (ch() == '%') {
		if (nch() == '=') {
			rch(); return(AMOD);
		}
		return(MOD);
	}
	else if (ch() == '&') {
		if (nch() == '=') {
			rch(); return(AAND);
		}
		else if (nch() == '&') {
			rch(); return(LAND);
		}
		return(AND);
	}
	else if (ch() == '|') {
		if (nch() == '=') {
			rch(); return(AOR);
		}
		else if (nch() == '|') {
			rch(); return(LOR);
		}
		return(OR);
	}
	else if (ch() == '^') {
		if (nch() == '=') {
			rch(); return(AXOR);
		}
		return(XOR);
	}
	else if (ch() == '<') {
		if (nch() == '=') {
			rch(); return(LSEQ);
		}
		else if (nch() == '<') {
			rch();
			if (nch() == '=') {
				rch(); return(ASHL);
			}
			return(SHL);
		}
		return(LESS);
	}
	else if (ch() == '>') {
		if (nch() == '=') {
			rch(); return(GTEQ);
		}
		else if (nch() == '>') {
			rch();
			if (nch() == '=') {
				rch(); return(ASHR);
			}
			return(SHR);
		}
		return(GRTR);
	}
	else if (ch() == '#') {
		if (nch() == '(') {
			rch(); return(LBRACE);
		}
		else if (nch() == ')') {
			rch(); return(RBRACE);
		}
		if (nch() == '<') {
			rch(); return(LBRACK);
		}
		else if (nch() == '>') {
			rch(); return(RBRACK);
		}
		else {
			return(HASH);
		}
	}
	else if (ch() == '{') {
		if (nch() == '{') {
			rch();
			if (nch() == '=') {
				rch(); return(AROL);
			}
			return(ROL);
		}
		return(LBRACE);
	}
	else if (ch() == '}') {
		if (nch() == '}') {
			rch();
			if (nch() == '=') {
				rch(); return(AROR);
			}
			return(ROR);
		}
		return(RBRACE);
	}

	else if (isalpha(ch()) || ch() == '_') {
		int	i = 0;

		while (isalpha(ch()) || ch() == '_' || isdigit(ch())) {
			if (i == TOKLEN) error(ETOKL, FATAL);
			if (i < TOKLEN-1) Strval[i++] = ch();
			rch();
		}
		Strval[i] = 0;
		back();

		if ((i = keyword(Strval)) != FUCKUP)
			return(i);

		Strval[NAMELEN] = 0;
		return(SYMBOL);
	}

	else if (ch() == '\'') {
		rch();
		Value = rche();
		if (ch() != '\'') error(ETERM, ERROR);
		return(NUMBER);
	}

	else if (isdigit(ch())) {
		Value = scanval();
		return(NUMBER);
	}

	else if (ch() == '"') {
		int	i = 0;

		rch();
		while (ch() != '"') {
			if (ch() == '\n') error(ETERM, ERROR);
			if (i == TOKLEN) error(ETOKL, FATAL);
			Strval[i++] = rche();
		}
		Strval[i] = 0;
		return(STRING);
	}

	else {
		error(ESCAN, FATAL);
	}

	return(FUCKUP);
}


/*
 *	NEXT()
 *	Purpose:	Read the next token from the input stream.
 *	Receives:	-
 *	Returns:	The token read.
 */
next() {
	return (Token = scan());
}
