/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	UTILITIES
 *
 *	Miscellaneous functions.
 *
 *	Public functions:
 *		NEWNODE()	allocate a node
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "token.h"
#include "glue.h"

#include "proto.h"


static int	Label;				/* Next Label to use */

static struct node	*Freelist;		/* Free list of node pool */
static struct node	*Usedlist,		/* List of used nodes */
			*Lastused;		/* Last used node */


/*
 *	UT_INIT()
 *	Purpose:	Initialize the utility routines.
 *	Receives:	NNOD	- The number of nodes to allocate
 *	Returns:	-
 */
void ut_init(nnod) {
	Label = 0;
	if ((Nodes = (struct node *) malloc(nnod*sizeof(struct node)))
	== NULL)
		error(EMEM, FATAL);
	Nsize = nnod;
	initnodes();
}


/*
 *	NEWNODE()
 *	Purpose:	Allocate a node, perform garbage collection.
 *	Receives:	TOKEN	- The desired type.
 *	Returns:	A pointer to the allocated node
 */
struct node *newnode(token) {
	struct node	*new;

	if (Freelist == NULL) {
		int	i;

		for (i=0; i<Nsize; i++) {
			if (Nodes[i].token == UNUSEDNODE) {
				Nodes[i].token = FREENODE;
				Nodes[i].nlink = Freelist;
				Freelist = &Nodes[i];
			}
		}
		if (Freelist == NULL) error(ENODES, FATAL);
	}

	new = Freelist;
	Freelist = Freelist->nlink;
	new->token = token;
	new->left = new->right = NULL;
	new->nlink = Usedlist;
	Usedlist = new;
	if (!Lastused) Lastused = new;
	return(new);
}


/*
 *	FREENODE()
 *	Purpose:	Release an allocated node.
 *	Receives:	NODE	- A pointer to the node.
 *	Returns:	-
 */
void freenode(node)
struct node	*node;
{
	node->token = UNUSEDNODE;
}


/*
 *	CLEARNODES()
 *	Purpose:	Release all used nodes in the node pool.
 *	Receives:	-
 *	Returns:	-
 */
void clearnodes() {
	struct node	*np;

	if (!Lastused) return;
	for (np=Usedlist; np; np = np->nlink)
		np->token = FREENODE;
	Lastused->nlink = Freelist;
	Freelist = Usedlist;
	Lastused = Usedlist = NULL;
}


/*
 *	INITNODES()
 *	Purpose:	Initialize the node pool.
 *	Receives:	-
 *	Returns:	-
 */
void initnodes() {
	int	i;

	for (i=0; i<Nsize; i++) {
		Nodes[i].nlink = &Nodes[i+1];
		Nodes[i].token = FREENODE;
	}
	Nodes[i-1].nlink = NULL;
	Freelist = Nodes;
	Lastused = Usedlist = NULL;
}


/*
 *	STRSAVE(S)
 *	Purpose:	Allocate string space, and copy the given string
 *			to the allocated space.
 *	Receives:	S	- The string to save
 *	Returns:	A pointer to the copy S.
 */
char *strsave(s)
char	*s;
{
	char	*new;

	if ((new = (char *) malloc(strlen(s)+1)) == NULL)
		error(EMEM, FATAL);
	strcpy(new, s);
	return(new);
}


/*
 *	NEWLABEL()
 *	Purpose:	Generate a unique label.
 *	Receives:	-
 *	Returns:	A new label.
 */
newlabel() {
	if (Label > MAXLBL) error(ELABEL, FATAL);
	return(Label++);
}


/*
 *	MATCH(TOK, WHAT, CLASS)
 *	Purpose:	Match a token, and generate a syntax error
 *			on failure.
 *	Receives:	TOK	- required token.
 *			WHAT	- verbose description of TOK
 *			CLASS	- class of error when the match
 *				fails.
 *	Returns:	-
 */
void match(tok, what, class)
int	tok;
char	*what;
int	class;
{
	if (tok == Token)
		next();
	else
		expect(what, class);
}


/*
 *	LPAREN()
 *	Purpose:	Match a left paren [(].
 *	Receives:	-
 *	Returns:	-
 */
void lparen() {
	match(LPAREN, "'('", ERROR);
}


/*
 *	RPAREN()
 *	Purpose:	Match a right paren [)].
 *	Receives:	-
 *	Returns:	-
 */
void rparen() {
	match(RPAREN, "')'", ERROR);
}


/*
 *	LBRACE()
 *	Purpose:	Match a left brace ({).
 *	Receives:	-
 *	Returns:	-
 */
void lbrace() {
	match(LBRACE, "'{'", ERROR);
}


/*
 *	RBRACE()
 *	Purpose:	Match a right brace (}).
 *	Receives:	-
 *	Returns:	-
 */
void rbrace() {
	match(RBRACE, "'}'", ERROR);
}


/*
 *	SEMI()
 *	Purpose:	Match a semicolon (;).
 *	Receives:	-
 *	Returns:	-
 */
void semi() {
	match(SEMI, "';'", ERROR);
}


/*
 *	SYMNAME()
 *	Purpose:	Match a symbol name.
 *	Receives:	-
 *	Returns:	-
 */
void symname() {
	match(SYMBOL, "symbol name", ERROR);
}


/*
 *	COMMA()
 *	Purpose:	Match a comma (,).
 *	Receives:	-
 *	Returns:	-
 */
void comma() {
	match(COMMA, "','", ERROR);
}


/*
 *	COLON()
 *	Purpose:	Match a colon (:).
 *	Receives:	-
 *	Returns:	-
 */
void colon() {
	match(COLON, "':'", ERROR);
}


/*
 *	LINK(ROOT, NEW)
 *	Purpose:	Link two nodes using a GLUE node.
 *	Receives:	ROOT	- the root node of the parse tree
 *			NEW	- the node to add
 *	Returns:	A new parse tree, where NEW is pasted to
 *			the head of ROOT (thru a GLUE node).
 */
struct node *link(root, new)
struct node	*root, *new;
{
	struct node	*t;

	t = newnode(GLUE);
	t->left = root;
	t->right = new;
	return(t);
}


/*
 *	PREPEND(NODE, TYPE)
 *	Purpose:	Create a new node and append it to the head of
 *			a node list.
 *	Receives:	NODE	- the head of the old list
 *			TYPE	- the type (token) of the new head
 *	Returns:	A new list with a type TYPE node as head.
 */
struct node *prepend(node, type)
struct node	*node;
int	type;
{
	struct node	*t;

	t = newnode(type);
	t->left = node;
	return(t);
}


/*
 *	CUT(ROOT)
 *	Purpose:	Removes a superflous GLUE node from the
 *			head of a parse tree (if any).
 *	Receives:	ROOT	- the parse tree to cut
 *	Returns:	The (possibly cut) parse tree.
 */
struct node *cut(root)
struct node	*root;
{
	struct node	*t;

	if (root->token == GLUE && root->left == NULL) {
		t = root;
		root = root->right;
		freenode(t);
	}
	return(root);
}


/*
 *	FETCH(NODE)
 *	Purpose:	Convert a temporary SYM node into a load operation.
 *			locals	-->	LL
 *			globals	-->	LG
 *			externs	-->	LX
 *	Receives:	NODE	- The SYM node (by reference)
 *	Returns:	Nothing, but modifies *NODE directly.
 */
void fetch(node)
struct node	**node;
{
	if ((*node)->token == G_SYM) {
		if ((*node)->v.sym->flags & EXTERN) {
			(*node)->token = G_LX;
			(*node)->v.val = findsx((*node)->v.sym->name);
		}
		else {
			(*node)->token = ((*node)->flags & GLOBAL ||
			(*node)->v.sym->flags & STATIC)?  G_LG: G_LL;
			(*node)->v.val = (*node)->v.sym->addr;
		}
	}
}


/*
 *	FINDSX(NAME)
 *	Purpose:	Find the symbol table index of a given symbol.
 *	Receives:	NAME	- the symbol to find.
 *	Returns:	The index of the symbol,
 *			or -1 if the symbol does not exist.
 */
findsx(name)
char	*name;
{
	int	i;

	for (i=0; i<St0; i++) {
		if (Symbols[i].flags & STRUCT)
			i += Symbols[i].size;
		else if (*name == *Symbols[i].name &&
		!strcmp(name, Symbols[i].name))
			return(i);
	}
	return(-1);
}


/*
 *	EMIT_SYM(FLAGS)
 *	Purpose:	Create a PUB/EXT symbol table (BCODE).
 *	Receives:	FLAGS	- desired flags
 *			When FLAGS & PUBLIC, build a PUB table.
 *			When FLAGS & EXTERN, build an EXT table.
 *	Returns:	-
 */
void emit_sym(flags) {
	int	i, nsym;

	t_emit(flags==PUBLIC? G_PUB: G_EXT);
	if (flags == EXTERN) flags |= REFERD;

	for (nsym=i=0; i<St0; i++) {
		if (Symbols[i].flags & STRUCT)
			i += Symbols[i].size;
		else if ((Symbols[i].flags & flags) == flags)
			nsym++;
	}
	n_emit(nsym);
	for (i=0; i<St0; i++) {
		if (Symbols[i].flags & STRUCT)
			i += Symbols[i].size;
		else if ((Symbols[i].flags & flags) == flags) {
			do_emit(Symbols[i].name);
			n_emit((flags & EXTERN)? i: Symbols[i].addr);
		}
	}
}
