/*
 *	MPP - a flexible macro preprocessor.
 *	Copyright (C) 1992, 1995 Nils M. Holm.
 *
 *	Usage: mpp [-h] [-p[char]] [-{kxy}[-]] [-s str] [-I dir] [file]
 *
 *	-h	print initial sync header
 *	-pc	set pasting character to c
 *	-k[-]	kill quotes: on [off] 			(off)
 *	-x[-]	expansion character required: on [off]	(off)
 *	-y[-]	case ignored: on [off]			(on)
 *	-f file	include profile `file'
 *	-s str	set sync string to str
 *	-I dir	set include directory to dir
 *	file	file to process (none means stdin)
 *
 *	MPP is FREE SOFTWARE, and comes with NO WARRANTY.
 *	The author cannot be held liable for any disadvantages
 *	resulting from the use of it.
 */

#define BSLSUPP		/* define if MPP is used as BSL preprocessor */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>

#define DFL_SYNC	"# \"%f\" %l%n"
#define DFL_INCDIR	"/v/bsl/mac"
#define ARGMAGIC	'\177'
#define EXPANDCH	'#'
#define MAXNEST		128

#define PATHL		81

#define message(s)	write(SYSERR, s, strlen(s))
#define unget(c)	if (Source) Sbuf = c; else Ubuf = c
#define iswhite(c)	((c) == ' '|| (c) == '\t' || (c) == '\f')
#define isname(c) 	(isalpha(c) || (c) == '_' || \
			((Flags&F_REQXC) && c == EXPANDCH))

#define EBUFL	16384
#define	BUFL	1024
#define STRL	256
#define TOKL	128
#define MAXARGS	128
#define SYSIN	0
#define SYSOUT	1
#define SYSERR	2

struct macro {
	char	*name,
		*val;
	int	narg;
	struct macro	*next;
};

struct builtin {
	char	*name;
	void	(*func)();
};


char	*Inbuf, Outbuf[BUFL];
int	Outfile;
int	Inp, Outp;
int	Inlen, Infile;
int	Ubuf, Sbuf;
int	Lquote, Rquote, Qlevel, Paste;
char	Sync[TOKL];
struct macro	*Macdefs, *Udefs;
char	Incdir[PATHL];
char	File[PATHL];
int	Line;
char	*Source;
int	Flags;
int	Maxcycles;

#define	F_ICASE		1
#define F_KILLQ		2
#define F_REQXC		4

void	defmac(), setincdir(), doinclude(), setquote(), setpaste(),
	redefmac(), setflags(), setsync(), undefmac(), sendto();

struct builtin	Bfuncs[] = {
	{	"define",	defmac,		},
	{	"incdir",	setincdir	},
	{	"include",	doinclude	},
	{	"quotechar",	setquote	},
	{	"pastechar",	setpaste	},
	{	"redefine",	redefmac	},
	{	"sendto",	sendto		},
	{	"setflags",	setflags	},
	{	"setsync",	setsync		},
	{	"undefine",	undefmac	},
	{	NULL,		NULL		},
};


char *num(n) {
	static char	nstr[10];
	int	i = 10;

	nstr[--i] = 0;
	while (n) {
		nstr[--i] = n%10+'0';
		n /= 10;
	}
	return(nstr+i);
}


void error(msg, name)
char	*msg, *name;
{
	void	printc();

	printc(-1);
	message("mpp (");
	message(num(Line));
	message(") : ");
	message(msg);
	if (name) {
		message(": ");
		message(name);
	}
	message("\n");
	exit(1);
}


void nomem() {
	error("insufficient memory", NULL);
}


void badeof() {
	error("unexpected EOF", NULL);
}


void overflow() {
	error("macro expansion buffer overflow", NULL);
}


readc() {
	int	c;

	if (Source) {
		if (Sbuf != -1) {
			c = Sbuf;
			Sbuf = -1;
			return(c);
		}
		if (!(*Source)) return(-1);
		return(*Source++);
	}
	if (Ubuf != -1) {
		c = Ubuf;
		Ubuf = -1;
		return(c);
	}
	if (Inp >= Inlen) {
		if ((Inlen = read(Infile, Inbuf, BUFL)) < 0)
			error("file read error", NULL);
		Inp = 0;
	}
	if (Inlen < 1) return(-1);
	if (Inbuf[Inp] == '\n') ++Line;
	return(Inbuf[Inp++]);
}


void printc(c) {
	if (Outfile < 0) return;
	if ((Flags & F_KILLQ) && (c == Lquote || c == Rquote)) return;
	if (c == -1 || Outp >= BUFL) {
		write(Outfile, Outbuf, Outp);
		Outp = 0;
		if (c == -1) return;
	}
	Outbuf[Outp++] = c;
}


void prints(s)
char	*s;
{
	while (*s) printc(*s++);
}


int skip() {
	int	c;

	c = readc();
#ifdef BSLSUPP
	while (iswhite(c) || c == '\n') {
		if (c == '\n') printc('\n');
		c = readc();
	}
#else
	while (iswhite(c)) c = readc();
#endif /* !BSLSUPP */
	return(c);
}


void lparen() {
	if (skip() != '(') error("missing left paren", NULL);
}


void rparen() {
	if (skip() != ')') error("missing right paren", NULL);
}


void comma() {
	if (skip() != ',') error("missing comma", NULL);
}


char *getname() {
	static char	buf[TOKL+1];
	int	c, i = 0;

	c = skip();
	if (!isname(c)) error("name expected", NULL);
	while (((isname(c) || isdigit(c)) || c == '.') && i < TOKL) {
		buf[i++] = c;
		c = readc();
	}
	unget(c);
	buf[i] = 0;
	return(buf);
}


char *getstr() {
	static char	buf[TOKL+1];
	int	c, i = 0, pl = 1;
	int	ql = 0;

	c = skip();
	for (;;) {
		if (ql && c == Rquote) {
			--ql;
		}
		else if (c == Lquote) {
			++ql;
		}
		else if (!ql && c == '(') {
			++pl;
		}
		else if (!ql && c == ')') {
			if (--pl < 1) break;
		}
		if (c == -1) badeof();
		buf[i++] = c;
		c = readc();
	}
	unget(c);
	buf[i] = 0;
	return(buf);
}


char *strsave(s)
char	*s;
{
	char	*new;

	if (!s) return(NULL);
	if ((new = (char *) malloc(strlen(s)+1)) == NULL)
		nomem();
	strcpy(new, s);
	return(new);
}


int strequ(s1, s2)
char	*s1, *s2;
{
	if (Flags & F_ICASE)
		return(!stricmp(s1, s2));
	else
		return(!strcmp(s1, s2));
}


struct macro *newmac(name, val, narg)
char	*name, *val;
int	narg;
{
	struct macro	*new;

	if ((new = (struct macro *) malloc(sizeof(struct macro))) == NULL)
		nomem();
	new->name = strsave(name);
	new->val = strsave(val);
	new->next = Macdefs;
	new->narg = narg;
	Macdefs = new;
	return(new);
}


struct macro *findmac(name)
char	*name;
{
	struct macro	*mp;

	if (Flags & F_REQXC) {
		if (*name != EXPANDCH) return(NULL);
		name++;
	}
	for (mp=Macdefs; mp; mp = mp->next) {
		if (strequ(mp->name, name)) return(mp);
	}
	return(NULL);
}


void fnmac(s)
char	*s;
{
	static char	name[TOKL+1], value[BUFL];
	char	anames[STRL], *p;
	int	ap = 0;
	int	narg = 0;
	int	c, pl, j, l, i;
	int	ql = 0;

	strcpy(name, s);

	if ((c = skip()) != ')') {
		unget(c);
		for (;;) {
			p = getname();
			anames[ap] = strlen(p)+2;
			strcpy(anames+ap+1, p);
			ap += anames[ap];
			narg++;
			if ((c = skip()) != ',') break;
		}
		unget(c);
		rparen();
	}
	comma();
	pl = 1;
	ql = 0;
	unget(skip());
	i = 0;
	for (; pl;) {
		if ((c = readc()) == -1) badeof();
		if (!ql && isname(c)) {
			unget(c);
			p = getname();
			ap = 0;
			for (j=0; j<narg; j++) {
				if (strequ(anames+ap+1, p)) break;
				ap += anames[ap];
			}
			if (j < narg) {
				value[i++] = ARGMAGIC;
				value[i++] = j+1;
			}
			else {
				if ((l = strlen(p)) + i >= EBUFL)
					overflow();
				strcpy(value+i, p);
				i += l;
			}
		}
#ifdef BSLSUPP
		else if (c == '\n') {
			printc('\n');
		}
		else if (iswhite(c)) {
			value[i++] = ' ';
			unget(skip());
		}
#endif /* BSLSUPP */
		else {
			value[i++] = c;
		}
		if (ql && c == Rquote)
			--ql;
		else if (c == Lquote)
			++ql;
		if (c == '(') ++pl;
		if (c == ')') --pl;
	}
	value[--i] = 0;
	unget(c);
	newmac(name, value, narg);
}


void strmac(name)
char	*name;
{
	char	value[TOKL+1];
	int	i, c, pl = 1;

	comma();
	unget(skip());
	for (i=0; pl && i<TOKL; i++) {
		if ((c = value[i] = readc()) == -1)
			badeof();
		if (c == '(') ++pl;
		if (c == ')') --pl;
	}
	value[--i] = 0;
	newmac(name, value, -1);
	unget(c);
}


void defmac() {
	char	*name;
	int	c;

	lparen();
	name = getname();
	if ((c = skip()) == '(') {
		fnmac(name);
	}
	else {
		unget(c);
		strmac(name);
	}
	rparen();
}


void setincdir() {
	lparen();
	strcpy(Incdir, getstr());
	rparen();
}


void writesync() {
	char	*p;

	for (p=Sync; *p; p++) {
		if (*p == '%') {
			switch(*++p) {
				case 'f':	prints(File); break;
				case 'l':	prints(num(Line)); break;
				case 'n':	printc('\n'); break;
				default:	printc(*p);
			}
		}
		else {
			printc(*p);
		}
	}
}


void includefile(fname, fd)
char	*fname;
int	fd;
{
	void	process();
	int	oldp, oldl, oldfd, oldln;
	char	oldfile[PATHL], *oldbuf;

	oldbuf = Inbuf;
	oldp = Inp;
	oldl = Inlen;
	oldfd = Infile;
	oldln = Line;
	strcpy(oldfile, File);

	Source = NULL;
	Infile = fd;
	if ((Inbuf = (char *) malloc(BUFL)) == NULL) nomem();
	Inp = Inlen = 0;
	strcpy(File, fname);
	Line = 1;
	writesync();

	process();

	free(Inbuf);
	Inbuf = oldbuf;
	Inp = oldp;
	Inlen = oldl;
	Infile = oldfd;
	Line = oldln;
	strcpy(File, oldfile);
	Source = "";
	writesync();
	close(fd);
}


void doinclude() {
	static char	path[PATHL];
	int	fd;
	char	*fname;

	lparen();
	fname = getstr();
	rparen();
	strcpy(path, Incdir);
	strcat(path, "/");
	if (*fname == '~') {
		strcat(path, fname+1);
		fname = path;
	}
	else {
		strcat(path, fname);
	}
	if ((fd = open(fname, O_RDONLY)) < 0) {
		if ((fd = open(path, O_RDONLY)) < 0) {
			error("no such file", fname);
		}
	}
	includefile(fname, fd);
}


void setquote() {
	lparen();
	Lquote = skip();
	if (Lquote == ',')
		Lquote = 0;
	else
		comma();
	Rquote = skip();
	if (Rquote == ')')
		Rquote = 0;
	else
		rparen();
}


void setpaste() {
	lparen();
	if ((Paste = skip()) == ')')
		Paste = 0;
	else
		rparen();
}


void setflags() {
	char	*p;
	int	n = 0;

	lparen();
	p = getstr();
	while(*p) {
		n = n*8+*p-'0';
		p++;
	}
	Flags = n;
	rparen();
}


void setsync() {
	lparen();
	strcpy(Sync, getstr());
	rparen();
}


void undefmac() {
	char	*name;
	struct macro	*mp, **mpp;

	lparen();
	name = getname();
	rparen();
	mpp = &Macdefs;
	for (mp=Macdefs; mp; ) {
		if (strequ(mp->name, name)) {
			*mpp = mp->next;
			mp->next = Udefs;
			Udefs = mp;
			return;
		}
		mpp = &mp->next;
		mp = mp->next;
	}
}


void redefmac() {
	char	*name;
	struct macro	*mp, **mpp;

	lparen();
	name = getname();
	rparen();
	mpp = &Udefs;
	for (mp=Udefs; mp; ) {
		if (strequ(mp->name, name)) {
			*mpp = mp->next;
			mp->next = Macdefs;
			Macdefs = mp;
			return;
		}
		mpp = &mp->next;
		mp = mp->next;
	}
	error("macro not undefined", name);
}


void sendto() {
	char	*name;
	int	c, append = 0;
	int	fd;

	lparen();
	if ((c = skip()) == ')') {
		printc(-1);
		Outfile = fileno(stdout);
		return;
	}
	unget(c);
	name = getname();
	if (strequ(name, "NULL")) {
		printc(-1);
		Outfile = -1;
		return;
	}
	if ((c = skip()) != ')') {
		unget(c); comma();
		getstr();
		rparen();
		append = 1;
	}
	if (append) {
		if ((fd = open(name, O_WRONLY)) >= 0)
			lseek(fd, 0L, SEEK_END);
	}
	else {
		fd = creat(name, S_IREAD|S_IWRITE);
	}
	if (fd < 0) {
		error("cannot open output file", name);
	}
	printc(-1);
	Outfile = fd;
}


int isbuiltin(s)
char	*s;
{
	int	i;

	if (Flags & F_REQXC) s++;
	for (i=0; Bfuncs[i].name; i++) {
		if (strequ(Bfuncs[i].name, s)) {
			(*Bfuncs[i].func)();
			return(1);
		}
	}
	return(0);
}


char *expand(s)
char	*s;
{
	static char	srcbuf[EBUFL];
	static char	expbuf[EBUFL];
	static char	argbuf[BUFL];
	struct macro	*mp;
	char	*args[MAXARGS];
	char	*p, *q;
	int	j, c, l, have_args;
	int	again, ql, pl, narg, ap, nl;

	strcpy(srcbuf, s);
	if ((c = readc()) == '(') {
		strcat(srcbuf, "(");
		strcat(srcbuf, getstr());
		readc();
		strcat(srcbuf, ")");
	}
	else {
		unget(c);
	}

	nl = 0;
	for (again=1; again;) {
		if (++nl >= Maxcycles) {
			error("possibly cyclic definition", NULL);
		}
		again = 0;
		ql = 0;
		j = 0;
		Source = srcbuf;
nextch:
		while ((c = readc()) != -1) {
			if (ql && c == Rquote) {
				--ql;
			}
			else if (c == Lquote) {
				++ql;
			}
			if (!ql && isname(c)) {
				unget(c);
				p = getname();
				if ((mp = findmac(p)) == NULL) {
					strcpy(expbuf+j, p);
					j += strlen(p);
					goto nextch;
				}

				again = 1;
				if (isbuiltin(p)) goto nextch;

				if (mp->narg == -1) {
					strcpy(expbuf+j, mp->val);
					j += strlen(mp->val);
					goto nextch;
				}

				lparen();
				narg = 0;
				ap = 0;
				p = argbuf;
				have_args = 0;
				for (pl=1; pl;) {
					if ((c = readc()) == -1) badeof();
					if (ql && c==Rquote) {
						--ql;
					}
					else if (c==Lquote) {
						++ql;
					}
					if (c == '(') {
						++pl;
					}
					else if (c == ')') {
						if (--pl < 1) break;
					}
					have_args = 1;
					if (c == ',' && !ql && pl<2) {
						if (ap >= BUFL) overflow();
						args[narg++] = p;
						argbuf[ap++] = 0;
						p = argbuf+ap;
						unget(skip());
					}
					else {
						if (ap >= BUFL) overflow();
						argbuf[ap++] = c;
					}

				}
				if (have_args) {
					argbuf[ap] = 0;
					args[narg++] = p;
				}
				if (narg != mp->narg)
					error("wrong number of arguments",
					mp->name);
				for (p=mp->val; *p; p++) {
					if (*p == ARGMAGIC) {
						q = args[*++p-1];
						l = strlen(q);
						if (l+j >= EBUFL) overflow();
						strcpy(expbuf+j, q);
						j += l;
					}
					else if (*p == Paste)
						;
					else {
						if (j >= EBUFL) overflow();
						expbuf[j++] = *p;
					}
				}
			}
			else {
				expbuf[j++] = c;
			}
		}
		expbuf[j] = 0;
		Source = NULL;
		strcpy(srcbuf, expbuf);
	}

	return(expbuf);
}


void evaluate(c) {
	unget(c);
	prints(expand(getname()));
}


void process() {
	int	c;

	while ((c = readc()) != -1) {
		if (c == Rquote && Qlevel)
			Qlevel--;
		else if (c == Lquote)
			Qlevel++;
		if (!Qlevel && isname(c))
			evaluate(c);
		else
			printc(c);
	}
	printc(-1);
}


void init() {
	int	i;

	if ((Inbuf = (char *) malloc(BUFL)) == NULL) nomem();
	Inp = Outp = 0;
	Inlen = 0;
	Sbuf = Ubuf = -1;
	Lquote = '\"';
	Rquote = '\"';
	Paste = '\0';
	Qlevel = 0;
	strcpy(Sync, DFL_SYNC);
	strcpy(Incdir, DFL_INCDIR);
	Macdefs = Udefs = NULL;
	for (i=0; Bfuncs[i].name; i++) newmac(Bfuncs[i].name, NULL, 0);
	Line = 1;
	strcpy(File, "(stdin)");
	Infile = SYSIN;
	Outfile = SYSOUT;
	Source = 0;
	Flags = F_ICASE;
	Maxcycles = MAXNEST;
}


void usage() {
	char	*text = "usage: mpp [-h] [-p[char]] [-{kxy}[-]] [-c count] "
			"[-s str] [-I dir] [-f file]\n\t\t[file]\n";

	write(SYSOUT, text, strlen(text));
	exit(1);
}


char *strip(s)
char	*s;
{
	if (*s == '\"') {
		s[strlen(s)-1] = 0;
		return(s+1);
	}
	return(s);
}


main(argc, argv)
int	argc;
char	**argv;
{
	int	i, p, pfd;
	int	isync = 0;
	char	*profile = NULL;

	init();
	for (p=1; p<argc && *argv[p] == '-'; p++) {
		for (i=1; argv[p][i]; i++) switch(argv[p][i]) {
			case 'c':	if (argv[p][i+1])
						Maxcycles = atoi(argv[p]+i);
					else
						Maxcycles = atoi(argv[++p]);
					i = strlen(argv[p])-1;
					break;
			case 's':	if (argv[p][i+1])
						strcpy(Sync, strip(argv[p]+i));
					else
						strcpy(Sync, strip(argv[++p]));
					i = strlen(argv[p])-1;
					break;
			case 'p':	if (argv[p][i+1]) {
						i++; Paste = argv[p][i];
					}
					else
						Paste = 0;
					break;
			case 'k':	if (argv[p][i+1] == '-') {
						i++; Flags &= ~F_KILLQ;
					}
					else
						Flags |= F_KILLQ;
					break;
			case 'x':	if (argv[p][i+1] == '-') {
						i++; Flags &= ~F_REQXC;
					}
					else
						Flags |= F_REQXC;
					break;
			case 'y':	if (argv[p][i+1] == '-') {
						i++; Flags &= ~F_ICASE;
					}
					else
						Flags |= F_ICASE;
					break;
			case 'I':	if (argv[p][i+1])
						strcpy(Incdir,
						strip(argv[p]+i));
					else
						strcpy(Incdir,
						strip(argv[++p]));
					i = strlen(argv[p])-1;
					break;
			case 'h':	isync = 1; break;
			case 'f':	if (argv[p][i+1]) {
						profile = argv[p]+i;
					}
					else
						profile = argv[++p];
					i = strlen(argv[p])-1;
					break;
			default:	usage();
		}
	}
	if (p < argc) {
		if ((Infile = open(argv[p], O_RDONLY)) < 0) {
			Line = 0;
			error("no such file", argv[p]);
		}
		strcpy(File, argv[p]);
	}
	if (profile) {
		if ((pfd = open(profile, O_RDONLY)) < 0) {
			Line = 0;
			error("cannot open profile", profile);
		}
		i = *Sync; *Sync = 0;
		includefile(profile, pfd);
		*Sync = i;
		Source = NULL; Inlen = Inp = 0;
	}
	if (isync) writesync();
	process();
	if (Infile != SYSIN) close(Infile);
	return(0);
}
