//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//	THE CONSTANT EXPRESSION PARSER
//
//	Reads a constant expression from the input file,
//	and returns the value of the expression.
//
//	Public functions:
//		CEXPR()		Parse a constant expression
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( token.hdr)


extern	next, expect, findsym,
	undeclared, stcsize,
	xrparen, report;

forward	cexpr;


//	C_PRIMARY()
//	Purpose:	Evaluates a primary factor in a const expression
//			Handles numeric literals, grouping, and constant
//			symbols.
//	Receives:	-
//	Returns:	the value of the factor
//
c_primary() {
	var	sp, v;

	if (Token == NUMBER) {
		v = Tvalue; next();
	}
	else if (Token == LPAREN) {
		next();
		v = cexpr();
		xrparen();
	}
	else if (Token == SYMBOL) {
		if (!(sp = findsym(Strval, 0))) {
			if (!(sp = findsym(Strval, 1))) {
				sp = undeclared(Strval);
			}
		}
		next();
		if (!(sp.S_flags & RVAL)) expect("constant", ERROR);
		v = (sp.S_flags & STCT)-> stcsize(sp), sp.S_addr;
		sp.S_flags |= REFD;
	}
	else {
		expect("constant factor", ERROR);
	}

	return v;
}


//	C_FACTOR()
//	Purpose:	Evaluates a factor in a const expression.
//			Handles unary operators.
//			Unary operators may be nested (e.g. --+10).
//	Receives:	-
//	Returns:	the value of the factor
//
c_factor() {
	var	v;

	if (Token == LNOT) {
		next();
		v = c_factor()-> 0, -1;
	}
	else if (Token == NOT) {
		next();
		v = ~c_factor();
	}
	else if (Token == SUB) {
		next();
		v = -c_factor();
	}
	else if (Token == ADD) {
		next();
		v = c_factor();
	}
	else {
		v = c_primary();
	}

	return v;
}


//	C_TERM()
//	Purpose:	Evaluates a term in a const expression.
//			Handles the binary operators * / %.
//	Receives:	-
//	Returns:	the value of the term
//
c_term() {
	var	u1, u2;

	u1 = c_factor();
	while (1) {
		switch (Token) {
			case MUL:
				next();
				u1 *= c_factor();
				break;

			case DIV:
				next(); u2 = c_factor();
				u2-> (u1 /= u2), report(EDIV, ERROR);
				break;

			case MOD:
				next(); u2 = c_factor();
				u2-> (u1 %= u2), report(EDIV, ERROR);
				break;

			default:
				return u1;
		}
	}
}


//	C_SUM()
//	Purpose:	Evaluates a sum in a const expression.
//			Handles the binary operators + and -.
//	Receives:	-
//	Returns:	the value of the sum
//
c_sum() {
	var	v;

	v = c_term();
	while (1) {
		switch (Token) {
			case ADD:
				next(); v += c_term(); break;

			case SUB:
				next(); v -= c_term(); break;

			default:
				return v;
		}
	}
}


//	C_SH_ROT()
//	Purpose:	Evaluates a chain of shift/rotate operations in a
//			constant expression.
//			Handles the binary operators << >> {{ }}.
//	Receives:	-
//	Returns:	the value of the operation.
//
c_sh_rot() {
	var	v;

	v = c_sum();
	while (1) {
		switch (Token) {
			case SHL:
				next(); v <<= c_sum(); break;

			case SHR:
				next(); v >>= c_sum(); break;

			case ROL:
				next(); v {{= c_sum(); break;

			case ROR:
				next(); v }}= c_sum(); break;

			default:
				return v;
		}
	}
}


//	C_BINAND()
//	Purpose:	Evaluates a chain binary AND operations in a
//			constant expression (operator &).
//	Receives:	-
//	Returns:	the value of the operation.
//
c_binand() {
	var	v;

	v = c_sh_rot();
	while (Token == AND) {
		next();
		v &= c_sh_rot();
	}
	return v;
}


//	C_BINXOR()
//	Purpose:	Evaluates a chain binary XOR operations in a
//			constant expression (operator ^).
//	Receives:	-
//	Returns:	the value of the operation.
//
c_binxor() {
	var	v;

	v = c_binand();
	while (Token == XOR) {
		next();
		v ^= c_binand();
	}
	return v;
}


//	CEXPR()
//	Purpose:	Evaluates constant expression.
//			Handles the binary OR operator (|).
//			CEXPR() reads the constant expression thru
//			the SCANNER.
//	Receives:	-
//	Returns:	the value of the constant expression.
//
public cexpr() {
	var	v;

	v = c_binxor();
	while (Token == OR) {
		next();
		v |= c_binxor();
	}
	return v;
}
