//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE OPTIMIZER
//
//	The optimizer performs some transformations on the parse tree
//	to make the resulting code more compact, and faster.
//
//	Public functions:
//		OPTIMIZE()	Optimize the parse tree
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( glue.hdr)
#include( string.hdr)


//	HEAD(LI)
//	Purpose:	Find the head of a label node. Skip labels,
//			GLUE nodes, and empty UNV-instructions between
//			the label and the head.
//	Receives:	LI	- the pool index of the label
//	Returns:	The pool index of the head. When there is no
//			parent node, the head is LI.
//
head(li) {
	var	a = Nodes + li;
	var	i, n;

	ufor (i=0, Nsize, NODE) {
		n = Nodes+i;
		if (n.N_left == a && !n.N_right || n.N_right == a) {
			if (n.N_token == GLUE || n.N_token == G_LBL ||
			n.N_token == G_DRO ||
			n.N_token == G_UNV && !(n.N_val))
				return head(i);
			else
				return i;
		}
	}
	return li;
}


//	FOLLOW(SI)
//	Purpose:	Follows subsequent JMP's thru the node pool.
//			Used to find the destination of a jump chain.
//	Receives:	SI	- The pool index of the origin (source)
//	Returns:	The pool index of the final destination
//
follow(si) {
	var	i, l, ni;

	while (1) {
 		l = (Nodes+si).N_val;
		ufor (i=0, Nsize, NODE) {
			if ((Nodes+i).N_token == G_LBL &&
			(Nodes+i).N_val == l) {
				ni = head(i);
				break;
			}
		}
		if ((Nodes+ni).N_token != G_JMP) return si;
		si = ni;
	}
}


//	OPT_JUMPS()
//	Purpose:	Elimininates jump to jumps, and jumps to directly
//			following labels. Directly modifies the extended
//			parse tree.
//	Receives:	-
//	Returns:	-
//
opt_jumps() {
	var	f, tn;
	var	i, j, t, dl;

	ufor (i=0, Nsize, NODE) {
		t = (Nodes+i).N_token;
		if (t == G_JMP || t == G_JTR || t == G_JFA) {
			(Nodes+i).N_val = (Nodes+follow(i)).N_val;

			dl = (Nodes+i).N_val;
			ufor (j=0, Nsize, NODE) {
				if ((Nodes+j).N_token == G_LBL &&
				(Nodes+j).N_val == dl) {
					tn = Nodes+j;
					while (1) {
						f = tn.N_right;
						if (!f) f = tn.N_left;
						if (!(tn = f)) break;
						if (f.N_token == G_JMP &&
						f.N_val == dl) {
							f.N_token = GLUE;
							break;
						}
						else if (f.N_token != GLUE) {
							break;
						}
					}
				}
			}
		}
	}
}


//	OPT_BRANCHES()
//	Purpose:	Eliminates conditional branches with constant
//			conditions. Directly modifies the extended
//			parse tree.
//	Receives:	-
//	Returns:	-
//
opt_branches() {
	var	headp;
	var	i, t;

	ufor (i=0, Nsize, NODE) {
		t = (Nodes+i).N_token;
		if (t == G_JTR || t == G_JFA) {
			if ((Nodes+i).N_right)
				headp = Nodes+i+N_right;
			else
				headp = Nodes+i+N_left;
			if ((.headp).N_token == G_LN) {
				if ((.headp).N_val) {
					(Nodes+i).N_token = (t == G_JTR)->
					G_JMP, GLUE;
				}
				else {
					(Nodes+i).N_token = (t == G_JFA)->
					G_JMP, GLUE;
				}
				.headp = 0;
			}
		}
	}
}


//	RC_OPT(ROOT)
//	Purpose:	Removes superflous instructions from right sides
//			of sub-expressions (like +0, *1, &0).
//	Receives:	ROOT	- A pointer to the parse tree.
//	Returns:	A pointer to the optimized parse treeA.
//
rc_opt(root) {
	var	rc = root.N_right;
	var	log, v;

	if (rc.N_val == 0) {
		switch (root.N_token) {
			case G_ADD: case G_SUB:
			case G_SHL: case G_SHR:
			case G_ROL: case G_ROR:
			case G_XOR: case G_OR:
				root.N_token = GLUE;
				root.N_right = 0;
				break;

			case G_MUL: case G_AND:
				root.N_token = G_LN;
				root.N_val = 0;
				root.N_left = root.N_right = 0;
				break;

			case G_DIV:
				// divide by zero !!
				break;
		}
		return root;
	}

	if (rc.N_val == 1) {
		switch (root.N_token) {
			case G_MUL: case G_DIV:
				root.N_token = GLUE;
				root.N_right = 0;
				break;

			case G_MOD:
				root.N_token = G_LN;
				root.N_val = 0;
				root.N_left = root.N_right = 0;
				break;
		}
		return root;
	}

	if (rc.N_val == -1) {
		switch (root.N_token) {
			case G_AND:
				root.N_token = GLUE;
				root.N_right = 0;
				break;

			case G_XOR:
				root.N_token = G_COM;
				root.N_right = 0;
				break;

			case G_OR:
				root.N_token = G_LN;
				root.N_val = -1;
				root.N_left = root.N_right = 0;
				break;
		}
		return root;
	}

	v = 2; log = 1;
	while (log < _BITS_) {
		if (rc.N_val == v) break;
		v *= 2;
		log += 1;
	}
	if (rc.N_val == v) {
		switch (root.N_token) {
			case G_MUL:
				root.N_token = G_SHL;
				rc.N_val = log;
				break;

			case G_DIV:
				root.N_token = G_SHR;
				rc.N_val = log;
				break;
		}
	}

	return root;
}


//	LC_OPT(ROOT)
//	Purpose:	Removes superflous instructions from left sides
//			of sub-expressions (like 0+, 1*, 0&).
//	Receives:	ROOT	- A pointer to the parse tree.
//	Returns:	A pointer to the optimized parse treeA.
//
lc_opt(root) {
	var	lc = root.N_left;
	var	log, v;

	if (lc.N_val == 0) {
		switch (root.N_token) {
			case G_MUL: case G_AND:
			case G_SHL: case G_SHR:
			case G_ROL: case G_ROR:
				root.N_token = G_LN;
				root.N_val = 0;
				root.N_left = root.N_right = 0;
				break;

			case G_SUB:
				root.N_token = G_NEG;
				root.N_left = 0;
				break;

			case G_ADD: case G_XOR: case G_OR:
				root.N_token = GLUE;
				root.N_left = 0;
				break;
		}
		return root;
	}

	if (lc.N_val == 1) {
		switch (root.N_token) {
			case G_MUL:
				root.N_token = GLUE;
				root.N_left = 0;
				break;
		}
		return root;
	}

	if (lc.N_val == -1) {
		switch (root.N_token) {
			case G_AND:
				root.N_token = GLUE;
				root.N_left = 0;
				break;

			case G_XOR:
				root.N_token = G_COM;
				root.N_left = 0;
				break;

			case G_OR:
				root.N_token = G_LN;
				root.N_val = -1;
				root.N_left = root.N_right = 0;
				break;
		}
		return root;
	}

	v = 2; log = 1;
	while (log < _BITS_) {
		if (lc.N_val == v) break;
		v *= 2;
		log += 1;
	}
	if (lc.N_val == v) {
		switch (root.N_token) {
			case G_MUL:
				root.N_token = G_SHL;
				lc.N_val = log;
				break;
		}
	}

	return root;
}


//	OPT_OPER(ROOT)
//	Purpose:	Remove superflous operations from expressions.
//	Receives:	ROOT	- A pointer to the extended parse tree.
//	Returns:	A pointer to the optimized parse tree.
//
opt_oper(root) {
	var	lc, rc;

	if (!(lc = root.N_left)) return root;
	if (!(rc = root.N_right)) return root;

	if (lc.N_token == G_LN) root = lc_opt(root);
	if (rc.N_token == G_LN) root = rc_opt(root);

	return root;
}


//	OPT_CEXPR(ROOT)
//	Purpose:	Evaluate constant sub-expressions, and remove
//			superflous operations.
//	Receives:	ROOT	- A pointer to the extended parse tree.
//	Returns:	A pointer to the optimized tree.
//
opt_cexpr(root) {
	var	lc, rc;
	var	v, lv, rv;

	if (!root) return root;
	if (root.N_token == G_NUM) return root; // Don't follow N-lists!

	root.N_left = opt_cexpr(root.N_left);
	root.N_right = opt_cexpr(root.N_right);

	lc = root.N_left;
	if (rc = root.N_right) {
		if (lc.N_token != G_LN || rc.N_token != G_LN)
			return opt_oper(root);
	}
	else {
		if (lc.N_token != G_LN) return opt_oper(root);
	}

	lv = lc.N_val;
	rv = rc.N_val;

	switch(root.N_token) {
		case G_ADD:	v = lv + rv; break;
		case G_SUB:	v = lv - rv; break;
		case G_MUL:	v = lv * rv; break;
		case G_DIV:	v = lv / rv; break;
		case G_MOD:	v = lv % rv; break;
		case G_NEG:	v = -lv; break;

		case G_AND:	v = lv & rv; break;
		case G_OR:	v = lv | rv; break;
		case G_XOR:	v = lv ^ rv; break;
		case G_SHL:	v = lv << rv; break;
		case G_SHR:	v = lv >> rv; break;
		case G_ROL:	v = lv {{ rv; break;
		case G_ROR:	v = lv }} rv; break;
		case G_COM:	v = ~lv; break;

		case G_EQ:	v = (lv == rv); break;
		case G_NE:	v = (lv != rv); break;
		case G_LT:	v = (lv < rv); break;
		case G_GT:	v = (lv > rv); break;
		case G_LE:	v = (lv <= rv); break;
		case G_GE:	v = (lv >= rv); break;
		case G_ULT:	v = (lv U< rv); break;
		case G_UGT:	v = (lv U> rv); break;
		case G_ULE:	v = (lv U<= rv); break;
		case G_UGE:	v = (lv U>= rv); break;
		case G_NOT:	v = !lv; break;

		default:	return root;
	}

	root.N_token = G_LN;
	root.N_val = v;
	root.N_right = root.N_left = 0;

	return root;
}


//	OPTIMIZE(ROOT)
//	Purpose:	Optimizes the generated parse tree, using
//			the following steps:
//				1) Evaluation of constant expressions
//				2) Elimination of conditional branches
//				   with constant conditions
//				3) Elimination of jumps to jumps.
//	Receives:	ROOT	- A pointer to the extended parse tree.
//	Returns:	A pointer to the optimized tree.
//
public optimize(root) {
	root = opt_cexpr(root);
	opt_branches();
	opt_jumps();
	return root;
}
