//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE STATEMENT PARSER
//
//	Parses a complete program, and builds/emits an extended parse
//	tree for each function. Also emits data initialization sequences.
//
//	Public functions:
//		PA_INIT()	Initialize the parser
//		FUNCTION()	parse a function body, and build/emit
//				the generated code
//		PARSE()		Parse a program
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( scan.hdr)
#include( token.hdr)
#include( glue.hdr)
#include( string.hdr)


var	loop_start,			// Label at beg. of loop
	loop_end,			// Label at end of loop
	loop_level;			// Nesting level at
					// beginning of loop


extern	newnode, freenode, expr,
	link, cut, newlabel, declare,
	findsym, clearnodes, alloc,
	next, xlparen, xrparen,
	xlbrace, xrbrace, xcolon, xcomma,
	xsemi, xsymname, cexpr, match,
	expect, undeclared, findsx,
	d_init, clrlocals, emit,
	do_emit, t_emit,
	setopts, fndecl, fwddecl, stcdecl,
	report;

forward	statement, declarations;


//	PA_INIT()
//	Purpose:	Initialize the parser.
//	Receives:	-
//	Returns:	-
//
public pa_init() {
	loop_start = loop_end = BADLABEL;
	Lno = 1;
	Errcount = 0;
	Level = 0;
	Dtop = 0;
	Ftop = 1;
}


//	IF_STMT()
//	Purpose:	Parse an IF-ELSE-statement.
//	Receives:	-
//	Returns:	An extended parse tree.
//
if_stmt() {
	var	true, false;
	var	root, t;

	next();
	xlparen();
	root = newnode(G_LBL);
	root.N_val = false = newlabel();
	t = root.N_left = newnode(G_JFA);
	t.N_val = false;
	t.N_right = expr(1);
	xrparen();
	root.N_right = statement();

	if (Token == KELSE) {
		next();
		t = newnode(G_JMP);
		t.N_val = true = newlabel();
		t.N_right = root.N_right;
		root.N_right = t;
		t = newnode(G_LBL);
		t.N_val = true;
		t.N_left = root;
		root = t;
		root.N_right = statement();
	}

	return root;
}


//	HALT_STMT()
//	Purpose:	Parse a HALT-statement
//	Receives:	-
//	Returns:	An extended parse tree.
//
halt_stmt() {
	var	root;

	next();
	root = newnode(G_HLT);
	root.N_left = expr(0);
	return root;
}


//	ADDNODE(ROOT, TYPE)
//	Purpose:	Add a node to a parse tree. If the head is
//			a GLUE node, change it, else create a new head.
//	Receives:	ROOT	- the head of the subtree
//			TYPE	- the desired token for the head
//	Returns:	The new head.
//
addnode(root, type) {
	var	t;

	if (root.N_token != GLUE) {
		t = newnode(type);
		t.N_left = root;
		root = t;
	}
	else {
		root.N_token = type;
	}
	return root;
}


//	MKSWITCH(NCASE, DFLT, LBL, VAL)
//	Purpose:	Build an extended parse tree for a SWITCH-statement.
//	Receives:	NCASE	- Number of CASEs in the SWITCH
//			DFLT	- The DEFAULT label
//			LBL	- A pointer to the label array
//			VAL	- A pointer to the value array (case values)
//	Returns:	The parse tree
//
mkswitch(ncase, dflt, lbl, val) {
	var	root, t;
	var	i;

	root = newnode(G_SWI);
	root.N_val = ncase;
	t = newnode(G_NUM);
	t.N_val = Dtop;
	t.N_right = root;
	root = t;
	t = newnode(G_NUM);
	t.N_val = dflt;
	t.N_right = root;
	root = t;

	for (i=0, ncase) {
		t = newnode(G_NUM);
		t.N_val = val.i;
		t.N_right = root;
		root = t;
		t = newnode(G_NUM);
		t.N_val = lbl.i;
		t.N_right = root;
		root = t;
	}

	alloc((ncase+1)*2, 1);

	return root;
}


//	SWITCH_STMT()
//	Purpose:	Parse a SWITCH-statement.
//	Receives:	-
//	Returns:	An extended parse tree.
//
switch_stmt() {
	var	root, t, exp;
	var	lbl.MAXCASE,
		val.MAXCASE;
	var	ncase, i;
	var	n_start, n_end, n_level;
	var	dflt = BADLABEL;

	next();
	xlparen();
	exp = expr(1);
	xrparen();

	root = newnode(G_JMP);
	n_start = loop_start;
	n_end = loop_end;
	n_level = Level;
	loop_start = root.N_val = newlabel();
	loop_end = newlabel();
	loop_level = Ftop;

	xlbrace();

	ncase = 0;
	while (1) {
		if (Token == KCASE) {
			if (ncase >= MAXCASE) report(ENCASE, FATAL);
			next();
			root = addnode(root, G_LBL);	// Fixed??
			lbl.ncase = root.N_val = newlabel();
			val.ncase = cexpr();
			for (i=0, ncase) {
				if (val.i == val.ncase) {
					report(EDCASE, ERROR);
					break;
				}
			}
			ncase += 1;
			xcolon();
		}
		else if (Token == KDFLT) {
			if (dflt != BADLABEL) report(EDCASE, ERROR);
			next();
			xcolon();
			root = addnode(root, G_LBL);	// Fixed??
			dflt = root.N_val = newlabel();
		}
		else if (Token == RBRACE) {
			break;
		}
		else {
			t = statement();
			root = link(root, t);
		}
	}
	xrbrace();

	root = addnode(root, G_JMP);
	root.N_val = loop_end;

	root = addnode(root, G_LBL);
	root.N_val = loop_start;
	root = link(root, exp);

	if (dflt == BADLABEL) dflt = loop_end;
	root = link(root, mkswitch(ncase, dflt, lbl, val));
	root.N_token = G_LBL;
	root.N_val = loop_end;

	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	return root;
}


//	WHILE_STMT()
//	Purpose:	Parse a WHILE-statement
//	Receives:	-
//	Returns:	An extended parse tree.
//
while_stmt() {
	var	root, t;
	var	begin, end, again;
	var	n_start, n_end, n_level;

	next();
	xlparen();
	root = newnode(G_JTR);
	root.N_val = again = newlabel();
	t = root.N_left = newnode(G_LBL);
	root.N_left.N_val = begin = newlabel();
	root.N_right = expr(1);
	end = newlabel();
	xrparen();

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = begin;
	loop_end = end;
	loop_level = Ftop;

	t.N_left = statement();

	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	t = newnode(G_LBL);
	t.N_val = end;
	t.N_right = root;
	root = t;
	t = root.N_left = newnode(G_LBL);
	t.N_val = again;
	t.N_left = newnode(G_JMP);
	t.N_left.N_val = begin;

	return root;
}


//	UNTIL_STMT()
//	Purpose:	Parse an UNTIL-statement
//	Receives:	-
//	Returns:	An extended parse tree.
//
until_stmt() {
	var	root, t;
	var	begin, end;
	var	n_start, n_end, n_level;

	next();
	root = newnode(G_LBL);
	root.N_val = end = newlabel();
	root.N_left = newnode(G_LBL);
	root.N_left.N_val = begin = newlabel();
	t = root.N_right = newnode(G_JFA);
	t.N_val = begin;

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = begin;
	loop_end = end;
	loop_level = Ftop;

	t.N_left = statement();

	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	match(KUNTIL, "'until'", ERROR);
	xlparen();
	t.N_right = expr(1);
	xrparen();


	return root;
}


//	FOR_INIT(PSYM)
//	Purpose:	Parse the INIT part of a FOR-statement:
//			FOR (VAR=EXPR, EXPR, CEXPR)
//			     ~~~~~~~~
//	Receives:	PSYM	- A pointer to the used counting
//				variable (by reference).
//			BEGIN	- Label that marks the logical
//				beginning of the statement.
//	Returns:	A parse tree containing the init code.
//			A pointer to the used symbol (thru PSYM).
//
for_init(psym, begin) {
	var	root;
	var	sname.NAMELEN+1;

	bcopy(Strval, sname, NAMELEN);
	sname.NAMELEN = 0;
	xsymname();
	if (!(.psym = findsym(sname, 0))) {
		if (!(.psym = findsym(sname, 1))) {
			.psym = undeclared(sname);
		}
	}
	(.psym).S_flags |= REFD;
	match(MOV, "'='", ERROR);

	root = newnode(G_JMP);
	root.N_val = begin;
	if ((.psym).S_flags & EXT) {
		root.N_right = newnode(G_SX);
		root.N_right.N_val = findsx((.psym).S_name);
	}
	else {
		root.N_right = newnode(((.psym).S_flags & STAT)-> G_SG, G_SL);
		root.N_right.N_val = (.psym).S_addr;
	}
	root.N_left = expr(0);

	return root;
}


//	FOR_COND(SYM, BEGIN, END)
//	Purpose:	Parse the CONDITION part of a FOR-statement:
//			FOR (VAR=EXPR, EXPR, CEXPR)
//				       ~~~~
//	Receives:	SYM	- A pointer to the symbol table entry of
//				the used counter.
//			BEGIN	- The label at the start of the loop
//			END	- The label at the end of the loop
//	Returns:	An extended parse tree containing code to
//			check the condition.
//
for_cond(sp, begin, again) {
	var	root, t, t2;

	root = newnode(G_JTR);
	root.N_val = again;
	t = root.N_right = newnode(GLUE);	// [U]LT OR [U]GT ??
	if (sp.S_flags & EXT) {
		t.N_left = newnode(G_LX);
		t.N_left.N_val = findsx(sp.S_name);
	}
	else {
		t.N_left = newnode((sp.S_flags & STAT)-> G_LG, G_LL);
		t.N_left.N_val = sp.S_addr;
	}
	t2 = t.N_left;
	t2.N_left = newnode(G_LBL);
	t2.N_left.N_val = begin;
	t.N_right = expr(0);
	return root;
}


//	FOR_INCR(SYM, INCR)
//	Purpose:	Build code for the INCREMENT part of a FOR-statement:
//			FOR (VAR=EXPR, EXPR, CEXPR)
//					     ~~~~~
//	Receives:	SYM	- A pointer to the symbol table entry of
//				the used counter.
//			INCR	- the increment (may be negative to do
//				a decrement).
//	Returns:	An extended parse tree.
//
for_incr(sp, incr, cont) {
	var	root, t;
	var	sx;

	root = newnode(G_ADD);
	if (sp.S_flags & EXT) {
		root.N_left = newnode(G_LX);
		sx = root.N_left.N_val = findsx(sp.S_name);
	}
	else {
		root.N_left = newnode((sp.S_flags & STAT)-> G_LG, G_LL);
		root.N_left.N_val = sp.S_addr;
	}

	t = root.N_left;
	t.N_left = newnode(G_LBL);
	t.N_left.N_val = cont;
	root.N_right = newnode(G_LN);
	root.N_right.N_val = incr;

	if (sp.S_flags & EXT) {
		t = newnode(G_SX);
		t.N_val = sx;
	}
	else {
		t = newnode((sp.S_flags & STAT)-> G_SG, G_SL);
		t.N_val = sp.S_addr;
	}
	t.N_right = root;
	root = newnode(G_DRO);
	root.N_right = t;

	return root;
}


//	FOR_STMT()
//	Purpose:	Parse a FOR-statement
//	Receives:	SIGN	- Respect signs in comparisons
//				(zero for UFOR).
//	Returns:	An extended parse tree.
//
for_stmt(sign) {
	var	root, t, op;
	var	sp;
	var	begin, cont, again, end;
	var	incr;
	var	n_start, n_end, n_level;

	next();
	xlparen();
	root = newnode(G_LBL);
	begin = newlabel();
	cont = newlabel();
	again = newlabel();
	root.N_val = end = newlabel();
	root.N_left = for_init(@sp, begin);
	xcomma();

	t = root.N_right = for_cond(sp, begin, again);
	op = t.N_right;

	if (Token == COMMA) {
		next();
		incr = cexpr();
	}
	else {
		incr = 1;
	}

	t.N_left = for_incr(sp, incr, cont);
	t = t.N_left;
	t = t.N_left = newnode(GLUE);
	t.N_left = newnode(G_LBL);
	t.N_left.N_val = again;

	xrparen();

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = cont;
	loop_end = end;
	loop_level = Ftop;

	t.N_right = statement();

	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	if (incr < 0) {
		op.N_token = sign-> G_GT, G_UGT;
	}
	else {
		op.N_token = sign-> G_LT, G_ULT;
	}

	return root;
}


//	LOOP_STMT(LABEL)
//	Purpose:	Parse a LOOP/BREAK statement (with sematic analysis).
//	Receives:	LABEL	- The destination label:
//				loop_start for LOOP
//				loop_end for BREAK
//	Returns:	An extended parse tree.
//
loop_stmt(label) {
	var	root;

	next();
	if (label == BADLABEL) {
		report(EBREAK, ERROR);
		return(0);
	}
	else {
		root = newnode(G_JMP);
		root.N_val = label;
		root.N_left = newnode(G_UNV);
		root.N_left.N_val = Ftop-loop_level;
		return root;
	}
}


//	RETURN_STMT()
//	Purpose:	Parse a RETURN statement
//	Receives:	-
//	Returns:	An extended parse tree.
//
return_stmt() {
	var	root;

	next();
	root = newnode(G_RTN);
	root.N_right = newnode(G_UNV);
	root.N_right.N_val = Ftop-1;
	if (Token != SEMI) root.N_left = expr(1);
	return root;
}


//	SIMPLE()
//	Purpose:	Parse a simple (non-compound) statement.
//	Receives:	-
//	Returns:	An extended parse tree
//
simple() {
	var	root;

	switch (Token) {
		case RBRACE:	return (0);
		case SEMI:	root = newnode(GLUE); break;
		case KBREAK:	root = loop_stmt(loop_end); break;
		case KDO:	root = until_stmt(); break;
		case KFOR:	root = for_stmt(1); return(root);
		case KUFOR:	root = for_stmt(0); return(root);
		case KHALT:	root = halt_stmt(); break;
		case KIF:	root = if_stmt(); return(root);
		case KLOOP:	root = loop_stmt(loop_start); break;
		case KRETURN:	root = return_stmt(); break;
		case KSWITCH:	root = switch_stmt(); return(root);
		case KWHILE:	root = while_stmt(); return(root);
		default:	root = expr(1);
	}
	xsemi();
	return root;
}


//	COMPOUND()
//	Purpose:	Increase the nesting level, parse a compound statement,
//			clear locals, and decrease the level again.
//	Receives:	-
//	Returns:	An extended parse tree.
//
compound() {
	var	root, base, t;
	var	oftop;

	next();
	Level += 1;
	oftop = Ftop;
	root = newnode(G_UNV);
	base = newnode(G_DRO);
	base.N_left = declarations(0);
	while (1) {
		if (!(t = statement())) break;
		if (Token == ENDFILE) break;
		base = link(base, t);
	}
	base = cut(base);
	xrbrace();
	root.N_left = base;
	root.N_val = (Ftop - oftop);
	Ftop = oftop;
	clrlocals(Level); Level -= 1;

	return root;
}


//	STATEMENT()
//	Purpose:	Parse a statement (simple or compound).
//			Signal end of statement for the code generator
//			(by adding DRO to the parse tree).
//	Receives:	-
//	Returns:	An extended parse tree.
//
var	Lastln = 0;
statement() {
	var	root;

	root = newnode(G_DRO);
	switch (Token) {
		case LBRACE:	root.N_left = compound(); break;
		default:	if (O_linfo && Lno != Lastln) {
					root.N_left = newnode(G_LIN);
					root.N_left.N_val = Lno;
					Lastln = Lno;
					root.N_right = simple();
				}
				else {
					root.N_left = simple();
				}
	}
	if (!root.N_left) {
		freenode(root); root = 0;
	}
	return root;
}


//	FUNCTION()
//	Purpose:	Parse a function body, emit the generated
//			code, and clear the node pool.
//	Receives:	-
//	Returns:	-
//
public function(label, forwrd) {
	var	ptree, t;

	t = newnode(G_LBL);
	t.N_val = label;
	if (!forwrd) t.N_left = d_init(CODE, label);

	ptree = newnode(G_RTN);
	ptree.N_left = newnode(G_FRM);
	ptree.N_right = statement();
	ptree = link(t, ptree);
	if (!Errcount) emit(ptree);
	clearnodes();
	clrlocals(0);
}


//	PARSE_OPTS()
//	Purpose:	Parse an OPTIONS statement, and set the
//			requested options.
//	Receives:	-
//	Returns:	-
//
parse_opts() {
	next();
	if (Token != STRING)
		expect("string literal", ERROR);
	else
		setopts(Strval);
	next();
	xsemi();
}


//	RESYNC()
//	Purpose:	Re-synchronize line numbers, and change input
//			file name.
//	Receives:	-
//	Returns:	-
//
resync() {
	next();
	if (Token != STRING) {
		expect("file name", ERROR);
	}
	else {
		if (.Inname == '(') {
			t_emit(G_NAM);
			do_emit(Strval);
		}
		strcpy(Inname, Strval);
	}
	next();
	if (Token != NUMBER)
		expect("line number", ERROR);
	else
		Lno = Tvalue-1;
	next();
}


//	DECLARATIONS(GLOBL)
//	Purpose:	Parse a chain of declarations.
//	Receives:	GLOBL	- Declarations are in global context.
//	Returns:	GLOBL	-> NULL
//			!GLOBL	-> An extended parse tree containing
//				the allocation/deallocation/initialization
//				code for the locals.
//
declarations(globl) {
	var	root = 0, t;

	while (1) {
		if (Token == KVAR) {
			next();
			t = declare(globl, 0);
		}
		else if (Token == KVALUE) {
			next();
			t = declare(globl, RVAL);
		}
		else if (Token == KEXTERN) {
			next();
			if (!globl) report(EDECL, ERROR);
			t = declare(globl, EXT);
		}
		else if (Token == KSTATIC) {
			next();
			if (Token == KVAR) next();
			t = declare(globl, STAT);
		}
		else if (Token == KPUBLIC) {
			next();
			if (Token == KVAR) {
				next();
				if (!globl) report(EDECL, ERROR);
				t = declare(1, STAT|PUB);
			}
			else {
				fndecl(PUB);
				t = 0;
			}
		}
		else if (Token == KFORWD) {
			next();
			t = fwddecl();
		}
		else if (Token == KSTRUCT) {
			next();
			stcdecl();
			t = 0;
		}
		else if (Token == KOPTS && globl) {
			parse_opts();
			t = 0;
		}
		else if (Token == HASH) {
			resync();
			t = 0;
		}
		else if (globl && Token == SYMBOL) {
			fndecl(0);
			t = 0;
		}
		else {
			if (Token == ENDFILE || !globl)
				return(globl-> 0, cut(root));
			expect("declaration", SEVERE);
		}

		if (globl) {
			emit(t); clearnodes();
		}
		else {
			root = link(root, t);
		}
	}
}


//	PARSE()
//	Purpose:	Parse a complete program.
//			PARSE() returns only if the EOF has been read
//			[in DECLARATIONS(TRUE)].
//	Receives:	-
//	Returns:	-
//
public parse() {
	next();
	declarations(1);
}
