///////////////////////////////////////////////////////////////////////////////
//
//	The BSL Compiler Kit
//	BSL Runtime Library: Regular Expression Matching
//	Copyright (C) 1995 Nils M. Holm
//
//	RXCOMP ( PATTERN, BUFFER )
//
//	PURPOSE:	Pre-compile a regular expression.
//	RECEIVES:	PATTERN	- The pattern to compile.
//			BUFFER	- Where to place the tokenized
//				pattern.
//	RETURNS:	The number of machine words occupied
//			by the compiled pattern.
//
// ---------------------------------------------------------------------------
//
//	RXMATCH ( PATTERN, STRING )
//
//	PURPOSE:	Compare a pattern with a string.
//	RECEIVES:	PATTERN	- The pattern to match.
//			STRING	- The string to compare with PATTERN.
//	RETURNS:	TRUE (-1) - STRING contains PATTERN
//			FALSE (0) - STRING does not contain PATTERN.
//
// ---------------------------------------------------------------------------
//
//	The used pattern matching algorithm is based upon the algoritm
//	used in the GREP utility provided by DECUS.
//	Unlike the DECUS pattern compiler, RXCOMP() never fails.
//	When a magic character is found in a strange place, it is
//	assumed to be non-magic. It will be matched literally in
//	this case.
//	Only a subset of the regular expression functionality has
//	been implemented:
//
//	.	matches any character
//
//	^	matches the beginning of the line
//		(only when placed at the beginning)
//
//	$	matches the end of the line
//		(only when it is the last character in the pattern)
//	
//	[...]	matches a class of characters
//		[ABC]	matches A or B or C
//		[A-Z]	matches all upper case characters
//		[z-a]	matches nothing (z > a)
//		[^a-z]	matches everything BUT a...z
//
//	-	when appended to a pattern, matches the
//		pattern optionally
//		[A-Za-z]-	matches a letter, if there is one
//
//	+	matches at least one occurence of a pattern
//		a+	matches a, aa, aaa, aaaa, but not the
//			empty string
//
//	*	matches zero or more occurences of a pattern
//		[0-9]*	matches all number, including the
//			empty string
//
//	UNIMPLEMENTED features are
//		- expression grouping	([A-Z][0-9])*
//		- forwarding		\(a*\)
//		- predefined groups	:x
//
//	EXAMPLE:
//
//		[A-Za-z_][A-Za-z0-9_]*
//
//		matches any BSL (or C) identifier:
//		first character must be alphabetic or '_'
//		following characters may be alphanumeric or '_'.
//

#include( blib.hdr)
#include( ../opt.hdr)


value	ANY	= 256,
	BOL	= 257,
	EOL	= 258,
	CLASS	= 259,
	XCLASS	= 260,
	SET	= 261,
	STAR	= 262,
	PLUS	= 263,
	DASH	= 264,
	END	= 0;


//
//	MKCLASS(BUFFER, PATP)	-- INTERNAL
//
//	PURPOSE:	Create a character class
//	RECEIVES:	BUFFER 	- Buffer to place the class in
//			PATP	- Pointer to pattern source
//	RETURNS:	The length of the class in machine words.
//			.PATP is set to the end of the pattern.
//
mkclass(buffer, patp) {
	var	pp = .patp,
		sp = .patp,
		bp = buffer;

	pp += 1;
	if (.pp == '^') {
		.bp = XCLASS;
		sp += 1;
		pp += 1;
	}
	else {
		.bp = CLASS;
	}
	bp += 2;
	while (.pp && .pp != ']') {
		if (.pp == '-') {
			if (pp == sp) {
				.bp = .pp;
			}
			else {
				.bp = .(bp-1);
				.(bp-1) = SET;
				bp += 1;
				if (.bp = .(pp+1)) pp += 1;
			}
		}
		else if (.pp == '\\') {
			if (.bp = .(pp+1)) pp += 1;
		}
		else {
			.bp = .pp;
		}
		bp += 1;
		pp += 1;
	}

	.patp = (.pp == ']')-> pp, pp-1;
	buffer.1 = bp-buffer-1;

	return bp - buffer;
}


public rxcomp(pattern, buffer) {
	var	pp = pattern,
		bp = buffer,
		ep = 0;

	while (.pp) {
		if (.pp != '*' && .pp != '+' && .pp != '-') ep = bp;
		switch (.pp) {
		case '.':
			.bp = ANY;
			break;

		case '^':
			if (pp == pattern)
				.bp = BOL;
			else
				.bp = '^';
			break;

		case '$':
			if (pp == pattern)
				.bp = '$';
			else
				.bp = EOL;
			break;

		case '[':
			bp += mkclass(bp, @pp)-1;
			break;

		case '*':
		case '+':
		case '-':
			if (!ep) {
				.bp = .pp;
			}
			else {
				var	p = bp;

				while (p U>= ep) {
					.(p+1) = .p;
					p -= 1;
				}
				.ep = ( (.pp == '*')-> STAR,
					(.pp == '+')-> PLUS,
					DASH );
 				bp += 1; .bp = END;
			}
			break;
		
		case '\\':
			if (.bp = .(pp+1)) pp += 1;
			break;

		default:
			.bp = .pp;
		}

		pp += 1;
		bp += 1;
	}
	.bp = END;

	return bp-buffer+1;
}


//
//	_RXMATCH(LINE0, LINE, PATTERN)	-- INTERNAL
//
//	PURPOSE:	The workhorse function of RXMATCH()
//	RECEIVES:	LINE0	- The complete line to match.
//			LINE	- The partial line to match
//				  (when recursing).
//			PATTERN	- The pattern to match.
//	RETURNS:	if LINE matches PATTERN, a pointer to
//			the first character AFTER the matched substring.
//			FALSE (0), if LINE does not match PATTERN.
//
_rxmatch(line0, line, pattern) {
	var	pp = pattern,
		lp = line;
	var	op, k, np, here;

	while ((op = .pp) != END) {
		pp += 1;
		switch (op) {
		case BOL:
			if (lp != line0) return 0;
			break;

		case EOL:
			if (.lp) return 0;
			break;

		case ANY:
			if (!(.lp)) return 0;
			lp += 1;
			break;

		case CLASS:
		case XCLASS:
			k = .pp;
			pp += 1;
			while (k -= 1) {
				if (.pp == SET) {
					pp += 2;
					k -= 2;
					if (.(pp-1) <= .lp && .lp <= .pp)
						break;
				}
				else {
					if (.pp == .lp) break;
				}
				pp += 1;
			}
			if (k && op == XCLASS)
				return 0;
			else if (!k && op == CLASS)
				return 0;
			pp += k;
			lp += 1;
			break;

		case DASH:
			np = _rxmatch(line0, lp, pp);
			while (.pp != END) pp += 1; pp += 1;
			if (np) lp = np;
			break;

		case PLUS:
			if (!(lp = _rxmatch(line0, lp, pp)))
				return 0;

		case STAR:
			here = lp;
			while (.lp && (np = _rxmatch(line0, lp, pp)) != 0)
				lp = np;
			while (.pp != END) pp += 1; pp += 1;
			while (lp U>= here) {
				if (np = _rxmatch(line0, lp, pp))
					return np;
				lp -= 1;
			}
			return 0;

		default:
			if (.lp != op) return 0;
			lp += 1;
			break;
		}
	}

	return lp;
}


public rxmatch(pattern, string) {
	var	i, k = 0;

	if (.pattern == BOL) {
		return !!_rxmatch(string, string, pattern);
	}
	while (string.k) k += 1;
	for (i=0, k) {
		if (_rxmatch(string+i, string+i, pattern)) return -1;
	}
	return (0);
}
