///////////////////////////////////////////////////////////////////////////////
//
//	BSL -- A BSL compiler shell.
//	Copyright (C) 1995 Nils M. Holm
//
//	Usage:	BSL [-?cfknpv] [-Ofile] [-Ldir] [-Idir] [-Bdir] [-Dname=val]
//		file [...]
//
//	BSL controls the single phases of the BSL compiler. It passed
//	each input file to the appropriate programs, depending on the
//	suffix of the file:
//
//		.bsl	preprocess, compile, resolve, load
//		.bsi	compile, resolve, load
//		.bco	resolve, load
//		.*	load
//
//		Step		Program		Input 	Output
//	1.	preprocess	BSLPP		.bsl	.bsi
//	2.	compile		BT		.bsi	.bco
//	3.	resolve		BCR		.bco	.bcr
//	4.	load		CG		.bcr	.exe
//
//	The steps 1-3 preserve existing output files. When step 1
//	finds an existing file with the .BSI suffix, it will abort
//	with a message like 'preserving existing file'.
//	Since the .EXE file is not removed by BSL for some clear
//	reasons, it will be silently overwritten by step 4, if it
//	exists.
//

#include(blib.hdr)
#include(../mem/mem.hdr)
#include(../ios/ios.hdr)
#include(../str/str.hdr)
#include(../sys/syscall.hdr)

#include(../opt.hdr)

value	POOLSIZE	= 8192;

var	PREPROCESSOR	= "bslpp.exe";
var	TRANSLATOR	= "bt.exe";
var	RESOLVER	= "bcr.exe";
var	CODE_GENERATOR	= "cg.exe";

var	Binary_path	= "c:/v/bsl/bin";
var	Header_path	= "c:/v/bsl/hdr";
var	Library_path	= "c:/v/bsl/lib";

var	Output_name	= 0;
var	Definitions	= 0;

var	Infd, Outfd, Savin, Savout;

var	O_stop = 3,
	O_verbose = 0,
	O_nexec = 0,
	O_force = 0,
	O_keep = 0;


quit(n) {
	stdclose();
	halt n;
}


terminate() {
	sputl(STDERR, "*** terminating\r\n");
	quit(1);
}


redirect(src, dest) {
	if ((Infd = sys_open(src, 0)) < 0) {
		swritef(STDERR, "BSL: cannot open: %s\r\n", src);
		quit(1);
	}
	if ((Savin = sys_dup(SYSIN)) < 0) {
		sputl(STDERR, "BSL: cannot duplicate SYSIN\r\n");
		sys_close(Infd);
		quit(1);
	}
	if (sys_dup2(Infd, SYSIN) < 0) {
		sys_close(Infd);
		sys_close(Savin);
		sputl(STDERR, "BSL: cannot redirect SYSIN\r\n");
		quit(1);
	}

	if ((Outfd = sys_create(dest)) < 0) {
		swritef(STDERR, "BSL: cannot create: %s\r\n", src);
		sys_close(Infd);
		sys_close(Savin);
		quit(1);
	}
	if ((Savout = sys_dup(SYSOUT)) < 0) {
		sputl(STDERR, "BSL: cannot duplicate SYSOUT\r\n");
		sys_close(Infd);
		sys_close(Savin);
		quit(1);
	}
	if (sys_dup2(Outfd, SYSOUT) < 0) {
		sputl(STDERR, "BSL: cannot redirect SYSOUT\r\n");
		sys_close(Infd);
		sys_close(Savin);
		sys_close(Outfd);
		sys_close(Savout);
		sys_remove(dest);
		quit(1);
	}
}


endred() {
	if ((sys_dup2(Savin, SYSIN) < 0) || (sys_dup2(Savout, SYSOUT) < 0)) {
		sputl(STDERR, "BSL: cannot restore SYSIO\r\n");
		quit(1);
	}
	sys_close(Savin);
	sys_close(Savout);
	sys_close(Infd);
	sys_close(Outfd);
}


execute(program, parameters, infile, outfile) {
	var	rc;

	if (O_verbose == 2) {
		swritef(STDERR, "%s %s <%s >%s\r\n",
		program, parameters-> parameters, "", infile, outfile);
		ssync(STDERR);
	}
	if (O_nexec) return 0;
	redirect(infile, outfile);
	if ((rc = sys_spawn(program, parameters)) < 0) {
		swritef(STDERR, "BSL: cannot spawn: %s\r\n", program);
		sys_remove(outfile);
		quit(1);
	}
	endred();
	return rc;
}


suffix(name, sfx) {
	var	ln, ls;

	if ((ln = strlen(name)) <= (ls = strlen(sfx))) return 0;
	return !strcmp(name+ln-ls, sfx);
}


chgname(name, sfx) {
	var	p;

	if ((p = strrchr(name, '.'))) .p = 0;
	strcat(name, sfx);
	return name;
}


preserve(name) {
	var	fd;

	if (O_nexec || O_force) return;
	if ((fd = sys_open(name, 0)) >= 0) {
		sys_close(fd);
		swritef(STDERR, "BSL: preserving existing file: %s\r\n",
		name);
		quit(1);
	}
}


exists(name) {
	var	fd;

	if (O_nexec) return;
	if ((fd = sys_open(name, 0)) < 0) {
		swritef(STDERR, "BSL: no such file: %s\r\n", name);
		quit(1);
	}
	sys_close(fd);
}


deltmp(name) {
	if (O_keep) return;
	if (O_verbose >= 2) {
		swritef(STDERR, "removing %s\r\n", name);
		ssync(STDERR);
	}
	if (O_nexec) return;
	if (sys_remove(name) < 0) {
		swritef(STDERR, "BSL: cannot remove: %s\r\n", name);
		quit(1);
	}
}


do_preprocess(name) {
	static var	oname . _PATHL_;
	var	args . 128,
		prog . 128;

	exists(name);
	strcpy(oname, name);
	chgname(oname, ".bsi");
	if (O_stop < 2 && Output_name) oname = Output_name;
	preserve(oname);
	strfmt(prog, "%s/%s", Binary_path, PREPROCESSOR);
	strfmt(args, "-I \"%s\" -P \"%s\"%s", Header_path, name,
	Definitions-> Definitions, "");
	if (execute(prog, args, name, oname)) {
		deltmp(oname);
		return 0;
	}
	return oname;
}


do_translate(name, delete) {
	static var	oname . _PATHL_;
	var	prog . 128;

	exists(name);
	strcpy(oname, name);
	chgname(oname, ".bco");
	if (O_stop < 3 && Output_name) oname = Output_name;
	preserve(oname);
	strfmt(prog, "%s/%s", Binary_path, TRANSLATOR);
	if (execute(prog, 0, name, oname)) {
		if (delete) deltmp(name);
		deltmp(oname);
		return 0;
	}
	if (delete) deltmp(name);
	return oname;
}


do_resload(name, delete) {
	static var	oname . _PATHL_;
	var	prog . 128,
		args . 128;

	exists(name);
	strcpy(oname, name);
	chgname(oname, ".bcr");
	preserve(oname);
	strfmt(prog, "%s/%s", Binary_path, RESOLVER);
	strfmt(args, "\"%s\"", Library_path);
	if (execute(prog, args, name, oname)) {
		if (delete) deltmp(name);
		deltmp(oname);
		return 0;
	}
	if (delete) deltmp(name);

	strcpy(name, oname);
	chgname(oname, ".exe");
	if (Output_name) oname = Output_name;
	strfmt(prog, "%s/%s", Binary_path, CODE_GENERATOR);
	if (execute(prog, args, name, oname)) {
		deltmp(name);
		deltmp(oname);
		return 0;
	}
	deltmp(name);
	return oname;
}


process(name) {
	var	prev = 0;

	if (O_verbose == 1) {
		swritef(STDERR, "%s:\r\n", name); ssync(STDERR);
	}

	if (suffix(name, "bsl")) {
		if (!(prev = do_preprocess(name))) terminate();
		if (O_stop < 2) return;
	}

	if (prev || suffix(name, "bsi")) {
		if (!(prev = do_translate(prev-> prev, name, prev)))
			terminate();
		if (O_stop < 3) return;
	}

	if (!do_resload(prev-> prev, name, prev)) terminate();
}


savearg(s) {
	var	new;

	if (!(new = malloc(strlen(s)+1))) {
		sputl(STDERR, "BSL: out of memory\r\n");
		quit(1);
	}
	strcpy(new, s);
	return new;
}


add_defs(s) {
	if (!Definitions) {
		if (!(Definitions = malloc(128))) {
			sputl(STDERR, "BSL: out of memory\r\n");
			quit(1);
		}
		.Definitions = 0;
	}
	strcat(Definitions, "\s-D");
	strcat(Definitions, s);
}


usage() {
	sputl(STDERR,
	"Usage: bsl [-?cfknpv] [-Ofile] [-Ldir] [-Idir] [-Bdir] ");
	sputl(STDERR, "[-Dname[=val]] file ...\r\n");
}


longusage() {
	sputl(STDERR, "BSL - Basic Systems Language  Version 1.0  ");
	sputl(STDERR, "Copyright (C) 1995 Nils M. Holm\r\n\n");
	usage();
	sputl(STDERR, "\r\n");
	sputl(STDERR, "-c\tcompile only, do not resolve or load\r\n");
	sputl(STDERR, "-n\tdo not execute any phases (debugging)\r\n");
	sputl(STDERR, "-p\tjust run the preprocessor\r\n");
	sputl(STDERR, "-v\tverbose - more v's give more information\r\n");
	sputl(STDERR, "-Ofile\tname the output file <file>\r\n");
	sputl(STDERR, "-Ldir\tLibrary is in directory <dir>\r\n");
	sputl(STDERR, "-Idir\tHeaders are in directory <dir>\r\n");
	sputl(STDERR, "-Bdir\tBinaries are in directory <dir>\r\n");
	sputl(STDERR, "-Dn=v\tDefine macro <n> with optional value <v>\r\n");
	sputl(STDERR, "\r\n");
}


start(argcount, argvec) {
	var	arena  . POOLSIZE;
	var	argbuf . 128;
	var	p, q, c;

	mminit(arena, POOLSIZE);
	stdopen();

	while (argcount) {
		unpack(.argvec, argbuf);
		if (.argbuf != '-') break;
		p = 1;
		while (argbuf.p) {
			switch (argbuf.p) {
				case '?':
					longusage(); quit(0);

				case 'c': case 'C':
					O_stop = 2; break;

				case 'f': case 'F':
					O_force = 1; break;

				case 'k': case 'K':
					O_keep = 1; break;

				case 'n': case 'N':
					O_nexec = 1; break;

				case 'p': case 'P':
					O_stop = 1; break;

				case 'v': case 'V':
					O_verbose += 1; break;

				case 'o': case 'O':
				case 'd': case 'D':
				case 'l': case 'L':
				case 'i': case 'I':
				case 'b': case 'B':
					c = argbuf.p;
					if (argbuf.(p+1)) {
						q = savearg(argbuf+p+1);
					}
					else {
						if (!(.(argvec += 1))) usage();
						argcount -= 1;
						unpack(.argvec, argbuf);
						q = savearg(argbuf);
					}
					p += strlen(argbuf+p)-1;
					switch (c) {
						case 'o': case 'O':
							Output_name = q;
							break;

						case 'l': case 'L':
							Library_path = q;
							break;

						case 'i': case 'I':
							Header_path = q;
							break;

						case 'b': case 'B':
							Binary_path = q;
							break;
						
						case 'd': case 'D':
							add_defs(q);
							break;
					}
					break;

				default:
					usage(); quit(1);
			}
			p += 1;
		}
		argvec += 1;
		argcount -= 1;
	}

	if (!argcount) usage();

	while (argcount) {
		unpack(.argvec, argbuf);
		process(argbuf);
		argvec += 1;
		argcount -= 1;
	}

	quit(0);
}
